from .pigpio_mocks import patch_pigpio_base, patch_pigpio_i2c, patch_pigpio_gpio, mock_i2c_hardware
from vent.io.devices import ADS1015, ADS1115, PigpioConnection
from vent.io.devices.sensors import Sensor, AnalogSensor, SFM3200, SimSensor
from secrets import token_bytes

import numpy as np
import pytest
import random


def test_sensor():
    """__________________________________________________________________________________________________________TEST #1
    Tests errors are thrown when you try to do stuff with the abstract base class
    """
    with pytest.raises(TypeError):
        s = Sensor()


def test_sim_sensor(patch_pigpio_i2c, mock_i2c_hardware):
    """__________________________________________________________________________________________________________TEST #2
    Tests the proper functioning of a simulated sensor. Not much to this one, really. Also checks that age sends back -1
        when update() has not been called
    """
    sim_sensor = SimSensor()
    assert sim_sensor.age() == -1
    n_iter = random.randint(100, 1000)
    for i in range(n_iter):
        result = sim_sensor.get()
        assert sim_sensor.low <= result <= sim_sensor.high
    sim_sensor.update()
    assert sim_sensor.age() > 0
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("ads1x15", [ADS1115, ADS1015])
@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(8)])
@pytest.mark.parametrize("pga", random.sample(ADS1115._CONFIG_VALUES[2], 4))
@pytest.mark.parametrize("mode", ['SINGLE', 'CONTINUOUS'])
def test_analog_sensor_single_read(patch_pigpio_i2c, mock_i2c_hardware, ads1x15, seed, pga, mode):
    """__________________________________________________________________________________________________________TEST #3
    Tests the proper functioning of an AnalogSensor with random, but plausible, calibration.

    - Performs a sequence of observations and verifies that each matches expectations and that the age of each reading
        properly reflects called updates
    """
    random.seed(seed)
    kwargs = {
        "MUX": random.choice(ads1x15._CONFIG_VALUES[1]),
        "PGA": pga,
        "MODE": mode,
        "DR": random.choice(ads1x15._CONFIG_VALUES[4]),
        'offset_voltage': round(random.uniform(-0.35, 0.35) * pga, 2),
        'output_span': round(random.uniform(.75, 1.0) * pga, 2),
        'conversion_factor': round(random.choice([-1, 1]) * random.uniform(1, 10), 2)
    }
    mock = mock_i2c_hardware(
        i2c_bus=1,
        i2c_address=ads1x15._DEFAULT_ADDRESS,
        n_registers=4,
        reg_values=[
            b'\x00\x00',
            b'\x85\x83'
        ]
    )
    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    ads = ads1x15(pig=pig)
    a_sensor = AnalogSensor(ads, **kwargs)
    n_iter = 250
    conversion_bytes = [token_bytes(2) for _ in range(n_iter)]
    raw_voltage = [int.from_bytes(cb, 'big', signed=True) * kwargs['PGA'] / 32767 for cb in conversion_bytes]
    expected = [
        kwargs['conversion_factor'] * (rv - kwargs['offset_voltage']) / kwargs['output_span'] for rv in raw_voltage
    ]
    for i in range(n_iter):
        pig.mock_i2c[1][mock['i2c_address']].write_mock_hardware_register(0, conversion_bytes[i])
        result = a_sensor.get()
        assert round(result, 9) == round(expected[i], 9)
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("ads1x15", [ADS1115, ADS1015])
@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(8)])
def test_analog_sensor_single_read_drop_kwargs(patch_pigpio_i2c, mock_i2c_hardware, ads1x15, seed):
    """__________________________________________________________________________________________________________TEST #4
    Same as the above test, except it will randomly drop kwargs to check that the default attributes are
        properly filled in and that the read functions normally.
        - Note: calibration values are stored seperately s.t. the later verification of the result can be calculated
         even without the store the default calibration for later conversion.
    """
    random.seed(seed)
    kwargs = {
        "MUX": random.choice(ads1x15._CONFIG_VALUES[1]),
        "PGA": random.choice(ads1x15._CONFIG_VALUES[2]),
        "MODE": random.choice(ads1x15._CONFIG_VALUES[3]),
        "DR": random.choice(ads1x15._CONFIG_VALUES[4]),
        'offset_voltage': 0,
        'output_span': 5,
        'conversion_factor': 10
    }
    kw_copy = kwargs.copy()
    for field in kw_copy:
        if random.getrandbits(1) and field is not "MUX":
            del kwargs[field]
            kw_copy[field] = dict(**ads1x15._DEFAULT_VALUES, **AnalogSensor._DEFAULT_CALIBRATION)[field]

    mock = mock_i2c_hardware(
        i2c_bus=1,
        i2c_address=ads1x15._DEFAULT_ADDRESS,
        n_registers=4,
        reg_values=[
            b'\x00\x00',
            b'\x85\x83'
        ]
    )
    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    ads = ads1x15(pig=pig)
    a_sensor = AnalogSensor(ads, **kwargs)
    n_iter = 250
    conversion_bytes = [token_bytes(2) for _ in range(n_iter)]
    raw_voltage = [int.from_bytes(cb, 'big', signed=True) * kw_copy['PGA'] / 32767 for cb in conversion_bytes]
    expected = [
        kw_copy['conversion_factor'] * (rv - kw_copy['offset_voltage']) / kw_copy['output_span'] for rv in raw_voltage
    ]
    for i in range(n_iter):
        pig.mock_i2c[1][mock['i2c_address']].write_mock_hardware_register(0, conversion_bytes[i])
        result = a_sensor.get()
        assert round(result, 9) == round(expected[i], 9)
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("ads1x15", [ADS1115, ADS1015])
@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(8)])
@pytest.mark.parametrize("pga", random.sample(ADS1115._CONFIG_VALUES[2], 4))
@pytest.mark.parametrize("mode", ['SINGLE', 'CONTINUOUS'])
def test_analog_sensor_kwarg_exceptions(patch_pigpio_i2c, mock_i2c_hardware, ads1x15, seed, pga, mode):
    """__________________________________________________________________________________________________________TEST #5
    Tests that an exception is properly raised when either:
        - a kwarg is passed that it unrecognized
        - MUX is not passed in kwargs
    """
    random.seed(seed)
    kwargs = {
        "MUX": random.choice(ads1x15._CONFIG_VALUES[1]),
        "PGA": pga,
        "MODE": mode,
        "DR": random.choice(ads1x15._CONFIG_VALUES[4]),
        'offset_voltage': round(random.uniform(-0.35, 0.35) * pga, 2),
        'output_span': round(random.uniform(.75, 1.0) * pga, 2),
        'conversion_factor': round(random.choice([-1, 1]) * random.uniform(1, 10), 2)
    }
    mock = mock_i2c_hardware(
        i2c_bus=1,
        i2c_address=ads1x15._DEFAULT_ADDRESS,
        n_registers=4,
        reg_values=[
            b'\x00\x00',
            b'\x85\x83'
        ]
    )
    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    ads = ads1x15(pig=pig)
    with pytest.raises(TypeError):
        bad_kwargs = kwargs.copy()
        bad_kwargs['bad_key'] = 'bad_value'
        AnalogSensor(ads, **bad_kwargs)
    with pytest.raises(TypeError):
        bad_kwargs = kwargs.copy()
        del bad_kwargs['MUX']
        AnalogSensor(ads, **bad_kwargs)
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("ads1x15", [ADS1115, ADS1015])
@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(32)])
def test_analog_sensor_average_read(patch_pigpio_i2c, mock_i2c_hardware, ads1x15, seed):
    """__________________________________________________________________________________________________________TEST #6
    Performs a several updates before averaging AnalogSensor.data to determine if the result matches expectations
        - Data loaded into the conversion register is gaussian normal about the expected value
    """
    random.seed(seed)
    pga = random.choice(ads1x15._CONFIG_VALUES[2])

    kwargs = {
        "MUX": random.choice(ads1x15._CONFIG_VALUES[1]),
        "PGA": pga,
        "MODE": random.choice(ads1x15._CONFIG_VALUES[3]),
        "DR": random.choice(ads1x15._CONFIG_VALUES[4]),
        'offset_voltage': round(random.uniform(-0.35, 0.35) * pga, 2),
        'output_span': round(random.uniform(.75, 1.0) * pga, 2),
        'conversion_factor': round(random.choice([-1, 1]) * random.uniform(1, 10), 2)
    }

    mock = mock_i2c_hardware(
        i2c_bus=1,
        i2c_address=ads1x15._DEFAULT_ADDRESS,
        n_registers=4,
        reg_values=[b'\x00\x00', b'\x85\x83']
    )

    def voltage(bytes_value):
        val = int.from_bytes(bytes_value, 'big', signed=True)
        result = val*kwargs['PGA'] / 32767
        return result

    def gaussian_bytes(mu):
        def _gauss():
            val = random.gauss(mu, mu/1000)
            result = int(round(val)).to_bytes(2, 'big')
            return result
        return _gauss

    n_iter = 1000
    target = random.randint(1000, 20000)

    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    ads = ads1x15(pig=pig)
    a_sensor = AnalogSensor(ads, **kwargs)
    a_sensor.maxlen_data = n_iter
    gauss_cannon = gaussian_bytes(target)
    bytes_data = [gauss_cannon() for _ in range(n_iter)]
    avg_voltage = np.mean([voltage(byt) for byt in bytes_data])
    expected_value = kwargs['conversion_factor'] * (avg_voltage - kwargs['offset_voltage']) / kwargs['output_span']

    for i in range(n_iter):
        pig.mock_i2c[1][mock['i2c_address']].write_mock_hardware_register(0, bytes_data[i])
        a_sensor.update()
    data = a_sensor.data
    result_0 = a_sensor.get(average=True)
    result_1 = data[:, 1].mean()
    assert len(data) == n_iter
    assert result_0 - result_1 < 0.1
    assert result_0 - expected_value < 0.1
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(4)])
def test_sfm_single_read(patch_pigpio_i2c, mock_i2c_hardware, seed):
    """__________________________________________________________________________________________________________TEST #7
    Tests the proper functioning of an SFM sensor. similar to single read of analog sensor, except starts with 4 bytes
        in the register and discards two (just like in real life)
    """
    random.seed(seed)
    mock = mock_i2c_hardware(
        i2c_bus=1,
        i2c_address=SFM3200._DEFAULT_ADDRESS,
        reg_values=[b'\x00\x00']
    )
    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    sfm = SFM3200(address=mock['i2c_address'], i2c_bus=mock['i2c_bus'], pig=pig)

    n_iter = random.randint(100, 1000)
    conversion_bytes = [token_bytes(4) for _ in range(n_iter)]
    raw_int = [int.from_bytes(cb[:2], 'big') for cb in conversion_bytes]
    expected = [(rv - sfm._FLOW_OFFSET) / sfm._FLOW_SCALE_FACTOR for rv in raw_int]
    for i in range(n_iter):
        pig.mock_i2c[1][mock['i2c_address']].write_mock_hardware_register(0, conversion_bytes[i])
        result = sfm.get()
        assert round(result, 10) == round(expected[i], 10)
    """__________________________________________________________________________________________________________
    """


# TODO: this is a unit test, need to add integration test
import random
import socket
import threading
import time
from unittest.mock import patch, Mock

import pytest

from vent.common import values
from vent.common.message import ControlSetting, SensorValues
from vent.alarm import AlarmSeverity, Alarm
from vent.common.values import ValueName
from vent.controller.control_module import ControlModuleBase
from vent.coordinator import rpc
from vent.coordinator.coordinator import get_coordinator


def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) != 0


class ControlModuleMock(ControlModuleBase):
    def __init__(self):
        super(ControlModuleMock, self).__init__()
        self.control_setting = {name: ControlSetting(name, -1, -1, -1, -1) for name in (ValueName.PIP,
                                                                                        ValueName.PIP_TIME,
                                                                                        ValueName.PEEP,
                                                                                        ValueName.BREATHS_PER_MINUTE,
                                                                                        ValueName.INSPIRATION_TIME_SEC)}
        self._running = threading.Event()

    def is_running(self):
        return self._running.is_set()

    def start(self):
        self._running.set()

    def get_sensors(self):
        self._running.wait()
        return SensorValues(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

    def get_control(self, control_setting_name: ValueName) -> ControlSetting:
        self._running.wait()
        return self.control_setting[control_setting_name]

    def set_control(self, control_setting: ControlSetting):
        self._running.wait()
        self.control_setting[control_setting.name] = control_setting

    def get_active_alarms(self):
        return {ValueName.PIP: Alarm(ValueName.PIP, True, AlarmSeverity.HIGH, time.time(), None)}

    def get_logged_alarms(self):
        return [Alarm(ValueName.PIP, False, AlarmSeverity.HIGH, time.time(), None)]


def mock_get_control_module(sim_mode):
    return ControlModuleMock()


@pytest.mark.parametrize("control_setting_name", values.controllable_values)
@patch('vent.controller.control_module.get_control_module', mock_get_control_module, Mock())
def test_local_coordinator(control_setting_name):
    coordinator = get_coordinator(single_process=True, sim_mode=True)
    coordinator.start()
    #while not coordinator.is_running():
    #    pass
    t = time.time()
    v = random.randint(10, 100)
    v_min = v - 5
    v_max = v + 5

    # TODO: add test for test reference

    c = ControlSetting(name=control_setting_name, value=v, min_value=v_min, max_value=v_max, timestamp=t)
    coordinator.set_control(c)

    c_read = coordinator.get_control(control_setting_name)
    assert c_read.name == c.name
    assert c_read.value == c.value
    assert c_read.min_value == c.min_value
    assert c_read.max_value == c.max_value
    assert c_read.timestamp == c.timestamp

@pytest.mark.timeout(10)
@pytest.mark.parametrize("control_setting_name", values.controllable_values)
@patch('vent.controller.control_module.get_control_module', mock_get_control_module, Mock())
def test_remote_coordinator(control_setting_name):
    # wait before
    #while not is_port_in_use(rpc.default_port):
    #    time.sleep(1)
    coordinator = get_coordinator(single_process=False, sim_mode=True)
    #TODO need to wait for rpc client start?
    #time.sleep(1)
    coordinator.start()
    #while not coordinator.is_running():
    #    pass
    t = time.time()
    v = random.randint(10, 100)
    v_min = v - 5
    v_max = v + 5

    # TODO: add test for test reference
    # TODO: test racing condition

    c = ControlSetting(name=control_setting_name, value=v, min_value=v_min, max_value=v_max, timestamp=t)
    coordinator.set_control(c)

    c_read = coordinator.get_control(control_setting_name)
    assert c_read.name == c.name
    assert c_read.value == c.value
    assert c_read.min_value == c.min_value
    assert c_read.max_value == c.max_value
    assert c_read.timestamp == c.timestamp


@pytest.mark.timeout(10)
def test_process_manager():
    # wait before
    #while not is_port_in_use(rpc.default_port):
    #    time.sleep(1)
    coordinator = get_coordinator(single_process=False, sim_mode=True)
    # TODO need to wait for rpc client start?
    #time.sleep(1)
    coordinator.start()
    #while not coordinator.is_running():
    #    pass

    assert coordinator.is_running() == True
    coordinator.process_manager.try_stop_process()
    assert coordinator.process_manager.child_pid is None

    try:
        coordinator.is_running()
        assert False
    except ConnectionRefusedError:
        pass
    except Exception:
        assert False

    coordinator.process_manager.start_process()

    #time.sleep(1)
    assert coordinator.process_manager.child_pid is not None
    assert coordinator.is_running() == False

    coordinator.process_manager.restart_process()

    #time.sleep(1)
    assert coordinator.process_manager.child_pid is not None
    assert coordinator.is_running() == False


def test_local_sensors():
    coordinator = get_coordinator(single_process=True, sim_mode=True)
    coordinator.start()
    #while not coordinator.is_running():
    #    pass

    sensor_values = coordinator.get_sensors()
    assert isinstance(sensor_values, SensorValues)

    values_dict = sensor_values.to_dict()
    # if all of the values are none, then this is just a blank object
    # FIXME: No longer relevant as blank SensorValues cannot be created anymore -- should ensure no values are None
    assert(not all([v is None for v in values_dict.values()]))
    for k, v in values_dict.items():
        assert isinstance(k, ValueName) or (k in sensor_values.additional_values)
        assert isinstance(v, int) or isinstance(v, float) or v is None

@pytest.mark.timeout(10)
def test_remote_sensors():
    # wait before
    #while not is_port_in_use(rpc.default_port):
    #    time.sleep(1)
    coordinator = get_coordinator(single_process=False, sim_mode=True)
    # TODO need to wait for rpc client start?
    #time.sleep(1)
    coordinator.start()
    #while not coordinator.is_running():
    #    pass

    sensor_values = coordinator.get_sensors()
    assert isinstance(sensor_values, SensorValues)
    values_dict = sensor_values.to_dict()
    # if all of the values are none, then this is just a blank object
    assert(not all([v is None for v in values_dict.values()]))

    for k, v in sensor_values.to_dict().items():
        assert isinstance(k, ValueName) or (k in sensor_values.additional_values)
        assert isinstance(v, int) or isinstance(v, float) or v is None


# def test_local_alarms():
#     coordinator = get_coordinator(single_process=True, sim_mode=True)
#     coordinator.start()
#     #while not coordinator.is_running():
#     #    pass
#
#     alarms = coordinator.get_active_alarms()
#     assert isinstance(alarms, dict)
#     for k, v in alarms.items():
#         assert isinstance(v, Alarm)
#
#     alarms = coordinator.get_logged_alarms()
#     assert isinstance(alarms, list)
#     for a in alarms:
#         assert isinstance(a, Alarm)

# @pytest.mark.timeout(10)
# def test_remote_alarms():
#     # wait before
#     #while not is_port_in_use(rpc.default_port):
#     #    time.sleep(1)
#     coordinator = get_coordinator(single_process=False, sim_mode=True)
#     # TODO need to wait for rpc client start?
#     #time.sleep(1)
#     coordinator.start()
#     #while not coordinator.is_running():
#     #    pass
#
#     alarms = coordinator.get_active_alarms()
#     assert isinstance(alarms, dict)
#     for k, v in alarms.items():
#         assert isinstance(v, Alarm)
#
#     alarms = coordinator.get_logged_alarms()
#     assert isinstance(alarms, list)
#     for a in alarms:
#         assert isinstance(a, Alarm)



from PySide2 import QtCore
from vent.alarm import AlarmSeverity, Alarm
import datetime
import time
from vent.common import values, message
import numpy as np

_ALARM_MANAGER_INSTANCE = None

def get_alarm_manager():

    if isinstance(globals()['_ALARM_MANAGER_INSTANCE'], AlarmManager):
        return globals()['_ALARM_MANAGER_INSTANCE']
    else:
        return AlarmManager()

class AlarmManager(QtCore.QObject):

    new_alarm = QtCore.Signal(Alarm)

    def __init__(self):
        super(AlarmManager, self).__init__()

        self.active_alarms = {}

    @QtCore.Slot(dict)
    def update_alarms(self, alarms):
        # FIXME: for now just forwarding
        self.active_alarms = alarms

        for alarm in alarms.values():
            alarm = self.parse_message(alarm)
            self.new_alarm.emit(alarm)

    @QtCore.Slot(tuple)
    def monitor_alarm(self, alarm):
        """
        Parse a tentative alarm from a monitor --
        we should have already gotten an alarm from the controller, so this
        largely serves as a double check.

        Doesn't use the :class:`~.message.Alarm` class because creating a new alarm increments
        the counter.

        Args:
            alarm (tuple): (monitor_name, monitor_value, timestamp)

        """
        # if alarm[0] in self.active_alarms.keys():
        #     return
        # else:
        #     # TODO: count these and raise an alarm that says the controller is out of sync
        #     new_alarm = Alarm(
        #         alarm_name=alarm[0],
        #         active = True,
        #         severity = AlarmSeverity.HIGH,
        #         start_time= alarm[2],
        #         alarm_end_time = None,
        #         value = alarm[1],
        #     )
        #     new_alarm = self.parse_message(new_alarm)
        #     self.new_alarm.emit(new_alarm)
        #     self.active_alarms[alarm[0]] = new_alarm
        pass



    def parse_message(self, alarm):
        """
        If an alarm doesn't have a ``message`` attr, make one for it.
        """
        if alarm.message is None:
            # make human readable time
            start_time = time.strftime(
                '%m/%d, %H:%M:%S',
                time.localtime(alarm.alarm_start_time)
            )

            if isinstance(alarm.alarm_name, values.ValueName):
                alarm_str = alarm.alarm_name.name
            else:
                alarm_str = str(alarm.alarm_name)


            if alarm.value is not None:
                # round to `digits` specified in value def
                round_digits = 1
                if alarm.alarm_name in values.CONTROL.keys():
                    round_digits = values.CONTROL[alarm.alarm_name]['decimals']
                elif alarm.alarm_name in values.SENSOR.keys():
                    round_digits = values.SENSOR[alarm.alarm_name]['decimals']

                value = np.round(alarm.value, decimals=round_digits)



                alarm.message = f"{alarm_str} went out of range at {start_time}, value was {value}"
            else:
                alarm.message = f"{alarm_str} went out of range at {start_time}"
        return alarm







from vent.io.devices.pins import Pin
from collections import deque
from functools import wraps
from random import getrandbits, choice
from secrets import token_bytes
from socket import error as socket_error

import pigpio
import pytest


@pytest.fixture()
def patch_pigpio_base(monkeypatch):
    """ monkeypatches the bare-minimum elements of pigpio to instantiate pigpio.pi() without communicating with the
    pigpio daemon (or hardware)"""

    class MockThread:
        def __init__(self, *args, **kwargs):
            pass

        def stop(self):
            pass

    class MockSocket:
        """ Bare-bones mock of socket.Socket with additional mocks of functions called by pigpio"""

        def __init__(self, *args, **kwargs):
            pass

        @staticmethod
        def close():
            pass

        def setsockopt(self, *args, **kwargs):
            pass

    class MockSockLock:
        """ Bare-bones mock of pigpio._socklock"""

        def __init__(self, *args, **kwargs):
            self.s = None
            self.l = None

    def mock_pigpio_command(*args, **kwargs):
        """ Make sure no commands get sent to the pigpio daemon"""
        pass

    def mock_create_connection(host, timeout):
        """ mock of socket.create_connection(). Returns a bare-bones mock socket"""
        return MockSocket

    def do_monkeypatch():
        monkeypatch.setattr("socket.create_connection", mock_create_connection)
        monkeypatch.setattr("pigpio._socklock", MockSockLock)
        monkeypatch.setattr("pigpio._callback_thread", MockThread)
        monkeypatch.setattr("pigpio._pigpio_command", mock_pigpio_command)
        monkeypatch.delattr("pigpio._pigpio_command_nolock")
        monkeypatch.delattr("pigpio._pigpio_command_ext")
        monkeypatch.delattr("pigpio._pigpio_command_ext_nolock")
       # monkeypatch.setattr("pigpio.pi.connected", 1, raising=False)

    do_monkeypatch()


@pytest.fixture()
def patch_bad_socket(monkeypatch):
    """ Monkeypatches socket.create_connection() to always throw an Exception as if things had gone poorly (instead of
            doing anything else)
    """
    def mock_create_bad_connection(host, timeout):
        """ mock of socket.create_connection(). Returns a bare-bones mock socket"""
        raise socket_error

    def do_monkeypatch():
        monkeypatch.setattr("socket.create_connection", mock_create_bad_connection)

    do_monkeypatch()


@pytest.fixture()
def patch_pigpio_i2c(patch_pigpio_base, monkeypatch):
    """ monkeypatches pigpio.pi() with a mock I2C interface that does not depend on the pigpio daemon"""

    def add_mock_hardware(self, device: MockHardwareDevice, i2c_address, i2c_bus):
        self.mock_i2c[i2c_bus][i2c_address] = device

    @mock_pigpio_errors
    def mock_i2c_open(self, i2c_bus, i2c_address):
        if i2c_bus in self.mock_i2c:
            handle = len(self.mock_handles)
            self.mock_handles[handle] = (i2c_bus, i2c_address)
            return handle
        else:
            return pigpio.PI_BAD_I2C_BUS

    @mock_pigpio_errors
    def mock_i2c_close(self, handle):
        if handle in self.mock_handles:
            (i2c_bus, i2c_address) = self.mock_handles[handle]
            del self.mock_i2c[i2c_bus][i2c_address]
            del self.mock_handles[handle]
            return 0
        else:
            return pigpio.PI_BAD_HANDLE

    @mock_pigpio_errors
    def mock_spi_open(self, channel, baudrate):
        """ Basically the same as I2C open except uses 'spi' as the i2c bus"""
        if channel in range(21):
            handle = len(self.mock_handles)
            self.mock_handles = {handle: ('spi', channel)}
            return handle
        else:
            return pigpio.PI_BAD_SPI_CHANNEL

    @mock_pigpio_errors
    def mock_spi_close(self, handle):
        """ Just calls I2C close."""
        result = self.i2c_close(handle)
        return result

    @mock_pigpio_errors
    def mock_i2c_read_device(self, handle, count):
        if handle in self.mock_handles:
            (bus, addr) = self.mock_handles[handle]
            return count, bytearray(self.mock_i2c[bus][addr].read_mock_hardware_device(count))
        else:
            return (pigpio.PI_BAD_HANDLE,)

    @mock_pigpio_errors
    def mock_i2c_read_i2c_block_data(self, handle, reg, count):
        if handle in self.mock_handles:
            (bus, addr) = self.mock_handles[handle]
            if count != 2:
                raise NotImplementedError
            return count, bytearray(self.mock_i2c[bus][addr].read_mock_hardware_register(reg, count))
        else:
            return (pigpio.PI_BAD_HANDLE,)

    @mock_pigpio_errors
    def mock_i2c_write_device(self, handle, data):
        if handle in self.mock_handles:
            (bus, addr) = self.mock_handles[handle]
            if type(data) is int:
                data = data.to_bytes(2, 'little')
            elif type(data) is bytes:
                pass
            else:
                raise NotImplementedError
            self.mock_i2c[bus][addr].write_mock_hardware_device(data)
            return 0
        else:
            return pigpio.PI_BAD_HANDLE

    @mock_pigpio_errors
    def mock_i2c_write_i2c_block_data(self, handle, reg, data):
        if handle in self.mock_handles:
            (bus, addr) = self.mock_handles[handle]
            if type(data) is int:
                data = data.to_bytes(2, 'little')
            elif type(data) is bytes:
                pass
            else:
                raise NotImplementedError
            self.mock_i2c[bus][addr].write_mock_hardware_register(reg, data)
            return 0
        else:
            return pigpio.PI_BAD_HANDLE

    def do_monkeypatch():
        monkeypatch.setattr("pigpio.pi.mock_i2c", {0: dict(), 1: dict(), 'spi': dict()}, raising=False)
        monkeypatch.setattr("pigpio.pi.mock_handles", dict(), raising=False)
        monkeypatch.setattr("pigpio.pi.add_mock_hardware", add_mock_hardware, raising=False)
        monkeypatch.setattr("pigpio.pi.i2c_open", mock_i2c_open)
        monkeypatch.setattr("pigpio.pi.i2c_close", mock_i2c_close)
        monkeypatch.setattr("pigpio.pi.i2c_read_device", mock_i2c_read_device)
        monkeypatch.setattr("pigpio.pi.i2c_read_i2c_block_data", mock_i2c_read_i2c_block_data)
        monkeypatch.setattr("pigpio.pi.i2c_write_device", mock_i2c_write_device)
        monkeypatch.setattr("pigpio.pi.i2c_write_i2c_block_data", mock_i2c_write_i2c_block_data)
        monkeypatch.setattr("pigpio.pi.spi_open", mock_spi_open)
        monkeypatch.setattr("pigpio.pi.spi_close", mock_spi_close)

    do_monkeypatch()


@pytest.fixture()
def patch_pigpio_gpio(patch_pigpio_base, monkeypatch):
    """ monkeypatches pigpio.pi with a mock GPIO interface.
    - Creates a new list "pigpio.pi.mock_pins" of MockPigpioPin objects, one for each RPi GPIO pin.
    - monkeypatches pigpio.pi.get_mode() to look at pigpio.pi.mock_pins for the mode of each pin instead of
        communicating with the pigpiod daemon.
    - monkeypatches pigpio.pi.set_mode() in a similar manner
    """

    class MockPigpioPin:
        def __init__(self, gpio: int):
            """ A simple object to mock a pigpio GPIO pin interface without communicating with the pigpiod daemon. It can
            be determined whether PWM is in use on the pin by checking if self.pwm_duty_cycle is None.

            Note: PWM frequency is weird. Basically, starting a hardware PWM with some frequency does not change the
                underlying (soft) PWM frequency. If the mode on the pin changes, the PWM frequency will revert to that
                soft frequency. However, while the hardware PWM is active, the hardware PWM frequency can be modified by
                calling set_PWM_frequency without triggering a change in mode - this DOES change the underlying soft PWM
                frequency, and will behave as though only soft frequencies are allowed. You can only set a PWM frequency
                outside of those allowed by soft_frequencies via starting a hardware PWM.

            Args:
                gpio (int): A number between 0 and 53
            """
            self.errors = dict(pigpio._errors)
            assert gpio in range(53)
            self._mode = choice([*Pin._PIGPIO_MODES.values()])
            self.gpio = gpio
            self.level = 0
            self.soft_pwm_frequency = 800
            self.hard_pwm_frequency = None
            self.pwm_duty_cycle = None

        @property
        def pwm_range(self):
            return 1000000 if self.mode == 4 else 255

        @property
        def pwm_frequency(self):
            return self.hard_pwm_frequency if self.mode == 4 else self.soft_pwm_frequency

        @pwm_frequency.setter
        def pwm_frequency(self, pwm_frequency):
            if self.mode == 4:
                self.hard_pwm_frequency = pwm_frequency
            self.soft_pwm_frequency = pwm_frequency

        @property
        def mode(self):
            return self._mode

        @mode.setter
        def mode(self, new_mode):
            """ Seems redundant with mock_pigpio_set_mode, but it isn't because mode can be set as a side effect of calling
            other methods, such as starting a PWM. Also, mode MUST be changed prior to setting a pwm_duty_cycle otherwise
            it will be unintentionally wiped
            """
            if new_mode != self.mode and self.pwm_duty_cycle is not None:
                self.pwm_duty_cycle = None
            self._mode = new_mode

    @mock_pigpio_bad_gpio_arg
    @mock_pigpio_errors
    def mock_pigpio_get_mode(self, gpio):
        """ Returns the mode if arg gpio is valid, otherwise mocks the pigpio errnum for PI_BAD_GPIO
        Args:
            self (pigpio.pi): A pigpio.pi or vent.io.devices.PigpioConnection instance
            gpio (int): A number between 0 and 53

        Returns:
            int: A number between 0 and 7
        """
        return self.mock_pins[gpio].mode

    @mock_pigpio_bad_gpio_arg
    @mock_pigpio_errors
    def mock_pigpio_set_mode(self, gpio, mode):
        """ Will stop a PWM if new mode != mode (via property setter), but not if new mode is the same as set mode
        Args:
            mode (int): A number between 0 and 7 (See: vent.io.devices.pins.Pin._PIGPIO_MODES)

        Returns:
            int: 0 if successful
        """
        if mode not in range(8):
            return self.mock_pins[gpio].errors[-4]
        else:
            self.mock_pins[gpio].mode = mode
            return 0

    @mock_pigpio_bad_gpio_arg
    @mock_pigpio_errors
    def mock_pigpio_read(self, gpio):
        """ Side effect: Will always set mode to 0 (which will always wipe a PWM)

        Returns:
            int: 0 if pin is pulled low, 1 if pin is pulled high
        """
        self.mock_pins[gpio].mode = 0
        return self.mock_pins[gpio].level

    @mock_pigpio_bad_gpio_arg
    @mock_pigpio_errors
    def mock_pigpio_write(self, gpio, level):
        """ Side effect: Will always wipe a PWM, and will set self.mode
        Args:
            level (int): 0 if pin is to be pulled low, 1 if pin is to be pulled high

        Returns:
            int: 0 if successful
        """
        if level not in range(2):
            return self.mock_pins[gpio].errors[-5]
        else:
            self.mock_pins[gpio].pwm_dutycycle = None
            self.mock_pins[gpio].mode = level
            self.mock_pins[gpio].level = level
            return 0

    @mock_pigpio_bad_user_gpio_arg
    @mock_pigpio_errors
    def mock_get_PWM_frequency(self, gpio):
        """ Note: Will return soft frequency if mode is not 4, regardless of whether a PWM is in use. If mode == 4, returns the
        hardware PWM frequency. This is handled via property getter for self.pwm_frequency

        Returns
            int: The frequency used for PWM on the GPIO, in Hz
        """
        result = self.mock_pins[gpio].pwm_frequency
        if result is not None:
            return result
        else:
            return self.mock_pins[gpio].errors[-92]

    @mock_pigpio_bad_user_gpio_arg
    @mock_pigpio_errors
    def mock_set_PWM_frequency(self, gpio, PWMfreq):
        """
        Args:
            PWMfreq (int): The frequency to be set on the pin.

        Returns:
            int: The frequency set on the GPIO. If PWMfreq not in soft_frequencies, it is the closest allowed frequency
        """
        if PWMfreq not in soft_frequencies:
            PWMfreq = min(soft_frequencies, key=lambda x: abs(x - PWMfreq))
        self.mock_pins[gpio].pwm_frequency = PWMfreq
        return PWMfreq

    @mock_pigpio_bad_user_gpio_arg
    @mock_pigpio_errors
    def mock_get_PWM_dutycycle(self, gpio):
        """
        Returns:
            int: The duty cycle used for the GPIO, out of the PWM range (default 255)
        """
        result = self.mock_pins[gpio].pwm_duty_cycle
        return result if result is not None else self.errors[-92]

    @mock_pigpio_bad_user_gpio_arg
    @mock_pigpio_errors
    def mock_set_PWM_dutycycle(self, gpio, PWMduty):
        """ Note: This will start a soft PWM and change pin mode if mode is not already 1. Must set self.mode prior
        to setting self.duty_cycle

        Args:
            PWMduty (int): the duty cycle to be use for the GPIO, must be in range(pwm_range+1)

        Returns:
            int: 0 if successful
        """

        if PWMduty in range(self.mock_pins[gpio].pwm_range + 1):
            self.mode = 1
            self.mock_pins[gpio].pwm_duty_cycle = PWMduty
            return 0
        else:
            return self.mock_pins[gpio].errors[-8]

    @mock_pigpio_bad_gpio_arg
    @mock_pigpio_errors
    def mock_get_PWM_range(self, gpio):
        """
        Returns:
            int: 255 unless hardware PWM then 1e6
        """
        return self.mock_pins[gpio].pwm_range

    @mock_pigpio_bad_gpio_arg
    @mock_pigpio_errors
    def mock_hardware_PWM(self, gpio, PWMfreq, PWMduty):
        """ Returns appropriate pigpio error if GPIO, PWMfreq, or PWMduty are not in their respective allowable ranges for
        pigpio hardware PWM.
        Note: must set self.pwm_duty_cycle prior to setting self.mode!

        Returns:
            int: 0 if successful
        """
        self.mock_pins[gpio].mode = 4
        if gpio not in (12, 13, 18, 19):
            return self.mock_pins[gpio].errors[-95]
        elif PWMfreq > 187500000 or PWMfreq < 0:
            return self.mock_pins[gpio].errors[-96]
        elif PWMduty not in range(self.mock_pins[gpio].pwm_range + 1):
            return self.mock_pins[gpio].errors[-97]
        else:
            self.mock_pins[gpio].hard_pwm_frequency = PWMfreq
            self.mock_pins[gpio].pwm_duty_cycle = PWMduty
            return 0

    def do_monkeypatch():
        monkeypatch.setattr("pigpio.pi.mock_pins", [MockPigpioPin(num) for num in range(53)], raising=False)
        monkeypatch.setattr("pigpio.pi.get_mode", mock_pigpio_get_mode)
        monkeypatch.setattr("pigpio.pi.set_mode", mock_pigpio_set_mode)
        monkeypatch.setattr("pigpio.pi.read", mock_pigpio_read)
        monkeypatch.setattr("pigpio.pi.write", mock_pigpio_write)
        monkeypatch.setattr("pigpio.pi.get_PWM_frequency", mock_get_PWM_frequency)
        monkeypatch.setattr("pigpio.pi.set_PWM_frequency", mock_set_PWM_frequency)
        monkeypatch.setattr("pigpio.pi.get_PWM_dutycycle", mock_get_PWM_dutycycle)
        monkeypatch.setattr("pigpio.pi.set_PWM_dutycycle", mock_set_PWM_dutycycle)
        monkeypatch.setattr("pigpio.pi.get_PWM_range", mock_get_PWM_range)
        monkeypatch.setattr("pigpio.pi.hardware_PWM", mock_hardware_PWM)

    do_monkeypatch()


@pytest.fixture()
def mock_i2c_hardware():
    """ Factory fixture for creating MockHardwareDevice objects
    """
    def make_mock_hardware(i2c_bus=None, i2c_address=None, n_registers=None, reg_values=None):
        """
        Args:
            i2c_bus (int): 0 or 1
            i2c_address: if None, will generate a random address. Otherwise, will pass through what it is given
            n_registers: If none, will be random.
            reg_values: #todo explain if n_registers > len(reg_values) remaining values filled in

        Returns:
            dict: keys = ('device', 'i2c_bus', 'i2c_address', 'expected') #todo explain what they are
        """
        i2c_bus = getrandbits(1) if i2c_bus is None else i2c_bus
        i2c_address = getrandbits(7) if i2c_address is None else i2c_address
        if reg_values is None:
            n_registers = getrandbits(5) if n_registers is None else n_registers
            reg_values = [token_bytes(2) for _ in range(n_registers)]
        else:
            if n_registers is None:
                pass
            elif n_registers > len(reg_values):
                for i in range(n_registers-len(reg_values)):
                    reg_values.append(token_bytes(2))
            elif n_registers < len(reg_values):
                raise ValueError("Cannot specify fewer registers than register values provided")
        device = MockHardwareDevice(*reg_values)
        return {'device': device, 'i2c_bus': i2c_bus, 'i2c_address': i2c_address, 'values': reg_values}
    return make_mock_hardware


def mock_pigpio_errors(func):
    @wraps(func)
    def mock_pigpio__u2i_exception(self, *args, **kwargs):
        value = func(self, *args, **kwargs)
        v = value if type(value) in (int, bool, float) else value[0]
        if v < 0:
            raise pigpio.error(pigpio.error_text(v))
        return value

    return mock_pigpio__u2i_exception


def mock_pigpio_bad_gpio_arg(func):
    @wraps(func)
    def check_args(self, gpio, *args, **kwargs):
        if gpio in range(53):
            result = func(self, gpio, *args, **kwargs)
        else:
            result = pigpio._errors[pigpio.PI_BAD_GPIO]
        return result

    return check_args


def mock_pigpio_bad_user_gpio_arg(func):
    @wraps(func)
    def check_args(self, gpio, *args, **kwargs):
        if gpio in range(31):
            result = func(self, gpio, *args, **kwargs)
        else:
            result = pigpio._errors[-2]
        return result

    return check_args


class MockHardwareDevice:
    def __init__(self, *args):
        """ A simple mock device with registers defined by kwargs. If only one register value is passed, device emulates
        a single register device that responds to read_device commands (such as the SFM3200). Retains past register
        values upon writing to a register for logging/assertion convenience.

        Register values are stored as raw bytes. Reading and writing to/from the registers simulates network-endian
            transactions by byteswapping.

        Args:
            *args: register_values of type bytes; one, per, register
        """
        self.last_register = None
        self.registers = []
        if args:
            for arg in args:
                if not isinstance(arg, bytes):
                    raise TypeError("Unknown register value of type {}".format(type(arg)))
                self.registers.append(deque())
                self.registers[-1].append(arg)
            if len(self.registers) == 1:
                self.last_register = 0
        else:
            TypeError("MockHardwareDevice needs at least one register to initialize")

    def read_mock_hardware_device(self, count=None):
        """Alias for read_mock_hardware_register(reg_num=None, count)"""
        return self.read_mock_hardware_register(reg_num=None, count=count)

    def write_mock_hardware_device(self, value):
        """ Alias for write_mock_hardware_register(reg_num=None, value)"""
        self.write_mock_hardware_register(reg_num=None, value=value)

    def read_mock_hardware_register(self, reg_num=None, count=None):
        """ Reads count bytes from a specific register

        Args:
            reg_num: The index of the register to read
            count (int): The number of bytes to read

        Returns:
            bytes: the register contents
        """
        if reg_num is None:
            if self.last_register is not None:
                reg_num = self.last_register
            else:
                raise RuntimeError("mock_i2c_device tried to access last register but none have been accessed yet")
        else:
            self.last_register = reg_num
        result = self.registers[reg_num][-1]
        return result if count is None else result[:count]

    def write_mock_hardware_register(self, reg_num, value):
        """ Writes value to register specified by reg_num

        Args:
            reg_num: The index of the register to write
            value (bytes): The stuff to write to the register
        """
        if type(value) is not bytes:
            raise TypeError("arg 'value' must be of type bytes")
        if reg_num is None:
            if self.last_register is not None:
                reg_num = self.last_register
            else:
                raise RuntimeError("mock_i2c_device tried to access last register but none have been accessed yet")
        else:
            self.last_register = reg_num
        self.last_register = reg_num
        self.registers[reg_num].append(value)


soft_frequencies = (8000, 4000, 2000, 1600, 1000, 800, 500, 400, 320, 250, 200, 160, 100, 80, 50, 40, 20, 10)


from vent.common import values
from copy import copy
from collections import OrderedDict as odict

class SensorValues:

    additional_values = ('timestamp', 'loop_counter', 'breath_count')
    def __init__(self, timestamp=None, loop_counter=None, breath_count=None, vals=None, **kwargs):
        """


        Args:
            timestamp (float): from time.time(). must be passed explicitly or as an entry in ``vals``
            loop_counter (int): number of control_module loops. must be passed explicitly or as an entry in ``vals``
            breath_count (int): number of breaths taken. must be passed explicitly or as an entry in ``vals``
            **kwargs: sensor readings, must be in :data:`vent.values.SENSOR.keys`
        """

        # if we were passed vals, we were given a dict of {ValueName: value}
        # allow this because python doesn't allow ** unpacking when the keys are not strings
        if vals is not None:
            for key, val in vals.items():
                if isinstance(key, values.ValueName):
                    kwargs[key.name] = val
                elif key in values.ValueName.__members__.keys():
                    kwargs[key] = val
                elif key in self.additional_values:
                    kwargs[key] = val

        # if we were called with a dictionary, make sure it came with
        # timestamp
        # loop_counter
        # breath_count
        self.timestamp = timestamp
        self.loop_counter = loop_counter
        self.breath_count = breath_count

        # if we were given None, try to get from kwargs.
        for val in self.additional_values:
            if getattr(self, val) is None:
                setattr(self, val, kwargs[val])

        # insist that we have all the rest of the vals
        assert(all([value.name in kwargs.keys() for value in values.SENSOR.keys()]))


        # assign kwargs as attributes,
        # don't allow any non-ValueName keys
        for key, value in kwargs.items():
            if (key in values.ValueName.__members__.keys()):
                setattr(self, key, copy(value))
            elif key in self.additional_values:
                continue
            else:
                raise KeyError(f'value {key} not declared in vent.values!!!')

    def to_dict(self):
        ret_dict = {
            valname: getattr(self,valname.name) for valname in values.SENSOR.keys()
        }

        ret_dict.update({
            k:getattr(self, k) for k in self.additional_values
        })

        return ret_dict

    def __getitem__(self, item):
        if item in values.ValueName:
            return getattr(self, item.name)
        elif item in values.ValueName.__members__.keys():
            return getattr(self, item)
        elif item.lower() in self.additional_values:
            return getattr(self, item.lower())
        else:
            raise KeyError(f'No such value as {item}')

    def __setitem__(self, key, value):
        if key in values.ValueName:
            return setattr(self, key.name, value)
        elif key in values.ValueName.__members__.keys():
            return setattr(self, key, value)
        elif key.lower() in self.additional_values:
            return setattr(self, key.lower(), value)
        else:
            raise KeyError(f'No such value as {key}')




class ControlSetting:
    def __init__(self, name, value, min_value, max_value, timestamp):
        """
        TODO: if enum is hard to use, we may just use a predefined set, e.g. {'PIP', 'PEEP', ...}
        :param name: enum belong to ValueName
        :param value:
        :param min_value:
        :param max_value:
        :param timestamp:
        """
        self.name = name
        self.value = value
        self.min_value = min_value
        self.max_value = max_value
        self.timestamp = timestamp


class Error:
    def __init__(self, errnum, err_str, timestamp):
        self.errnum = errnum
        self.err_str = err_str
        self.timestamp = timestamp





#!/usr/bin/env python
import argparse
import sys
import os
from vent import prefs
from vent.gui.main import launch_gui
from vent.coordinator.coordinator import get_coordinator


def parse_cmd_args():
    parser = argparse.ArgumentParser()
    # TODO: maybe we should add a mode without UI display, so this would only have command line interface?
    parser.add_argument('--simulation',
                        help='run as simulation using virtual sensors and actuators (default: False)',
                        action='store_true')
    parser.add_argument('--single_process',
                        help='running UI and coordinator within one process (default: False)',
                        action='store_true')
    return parser.parse_args()


def main():
    args = parse_cmd_args()
    coordinator = get_coordinator(single_process=args.single_process, sim_mode=args.simulation)
    app, gui = launch_gui(coordinator)
    sys.exit(app.exec_())


    # TODO: gui.main(ui_control_module)


if __name__ == '__main__':
    main()



""" A module for ventilator hardware device drivers
"""
from .base import (
    PigpioConnection,
    IODeviceBase,
    I2CDevice,
    SPIDevice,
    ADS1115,
    ADS1015,
    be16_to_native,
    native16_to_be
)



import numpy as np
import time
from PySide2 import QtWidgets, QtCore

from vent.gui import styles, mono_font
from vent.gui.widgets.components import RangeSlider
from vent.common import message


class Monitor(QtWidgets.QWidget):
    alarm = QtCore.Signal(tuple)
    limits_changed = QtCore.Signal(tuple)

    def __init__(self, value, update_period=styles.MONITOR_UPDATE_INTERVAL, enum_name = None):
        """

        Args:
            value (:class:`~vent.values.Value`):
            update_period (float): update period of monitor in s
        """
        super(Monitor, self).__init__()

        self.name = value.name
        self.units = value.units
        self.abs_range = value.abs_range
        self.safe_range = value.safe_range
        self.decimals = value.decimals
        self.update_period = update_period
        self.enum_name = enum_name

        self.value = None

        # whether we are currently styled as being in an alarm state
        self._alarm = False

        self.init_ui()

        self.timed_update()

    def init_ui(self):
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        #########
        # create widgets
        # make range slider
        self.range_slider = RangeSlider(self.abs_range, self.safe_range,
                                        decimals=self.decimals,
                                        orientation=QtCore.Qt.Orientation.Horizontal)

        # make comboboxes to display numerical value
        self.max_safe = QtWidgets.QDoubleSpinBox()
        self.max_safe.setDecimals(self.decimals)
        self.max_safe.setRange(self.abs_range[0], self.abs_range[1])
        self.max_safe.setSingleStep(10 ** (self.decimals * -1))
        self.max_safe.setValue(self.safe_range[1])

        self.min_safe = QtWidgets.QDoubleSpinBox()
        self.min_safe.setDecimals(self.decimals)
        self.min_safe.setRange(self.abs_range[0], self.abs_range[1])
        self.min_safe.setSingleStep(10 ** (self.decimals * -1))
        self.min_safe.setValue(self.safe_range[0])

        # labels to display values
        self.value_label = QtWidgets.QLabel()
        self.value_label.setStyleSheet(styles.DISPLAY_VALUE)
        self.value_label.setFont(mono_font())
        self.value_label.setAlignment(QtCore.Qt.AlignRight)
        self.value_label.setMargin(0)
        self.value_label.setContentsMargins(0,0,0,0)
        self.value_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                       QtWidgets.QSizePolicy.Expanding)

        self.name_label = QtWidgets.QLabel()
        self.name_label.setStyleSheet(styles.DISPLAY_NAME)
        self.name_label.setText(self.name)
        self.name_label.setAlignment(QtCore.Qt.AlignRight)


        self.units_label = QtWidgets.QLabel()
        self.units_label.setStyleSheet(styles.DISPLAY_UNITS)
        self.units_label.setText(self.units)
        self.units_label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignTop)
        self.units_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                         QtWidgets.QSizePolicy.Expanding)

        # toggle button to expand control
        self.toggle_button = QtWidgets.QToolButton(checkable=True,
                                                   checked=False)
        self.toggle_button.setStyleSheet(styles.DISPLAY_NAME)

        self.toggle_button.toggled.connect(self.toggle_control)
        self.toggle_button.setToolButtonStyle(
            QtCore.Qt.ToolButtonIconOnly
        )
        self.toggle_button.setSizePolicy(QtWidgets.QSizePolicy.Maximum,
                                         QtWidgets.QSizePolicy.Expanding)
        self.toggle_button.setArrowType(QtCore.Qt.RightArrow)

        #########
        # connect widgets

        # update boxes when slider changed
        self.range_slider.valueChanged.connect(self.update_boxes)

        # and vice versa
        self.min_safe.valueChanged.connect(self.range_slider.setLow)
        self.max_safe.valueChanged.connect(self.range_slider.setHigh)

        # and connect them all to a general limits_changed method
        # that also checks the alarm
        self.range_slider.valueChanged.connect(self._limits_changed)
        self.min_safe.valueChanged.connect(self._limits_changed)
        self.max_safe.valueChanged.connect(self._limits_changed)

        #########
        # layout widgets

        # first make label layout
        label_layout = QtWidgets.QGridLayout()
        label_layout.setContentsMargins(0,0,0,0)
        label_layout.addWidget(self.toggle_button, 0,0,2,1)
        label_layout.addWidget(self.value_label, 0,1,2,1)
        label_layout.addWidget(self.name_label, 0,2,1,1)
        label_layout.addWidget(self.units_label, 1,2,1,1)
        # label_layout.addStretch()
        self.layout.addLayout(label_layout, 5)

        # then combine sliders and boxes
        self.slider_layout = QtWidgets.QVBoxLayout()

        minmax_layout = QtWidgets.QHBoxLayout()
        minmax_layout.addWidget(QtWidgets.QLabel('Min:'))
        minmax_layout.addWidget(self.min_safe)
        minmax_layout.addStretch()
        minmax_layout.addWidget(QtWidgets.QLabel('Max:'))
        minmax_layout.addWidget(self.max_safe)
        self.slider_layout.addLayout(minmax_layout)

        self.slider_layout.addWidget(self.range_slider)

        # and wrap them in a widget so we can show/hide more robustly
        self.slider_frame = QtWidgets.QFrame()
        self.slider_frame.setLayout(self.slider_layout)
        self.slider_frame.setVisible(False)


        #self.layout.addLayout(self.slider_layout)



    @QtCore.Slot(bool)
    def toggle_control(self, state):
        if state == True:
            self.toggle_button.setArrowType(QtCore.Qt.DownArrow)
            self.layout.addWidget(self.slider_frame)
            self.slider_frame.setVisible(True)
            self.adjustSize()
        else:
            self.toggle_button.setArrowType(QtCore.Qt.LeftArrow)
            self.layout.removeWidget(self.slider_frame)
            self.slider_frame.setVisible(False)
            self.adjustSize()


    def update_boxes(self, new_values):
        self.min_safe.setValue(new_values[0])
        self.max_safe.setValue(new_values[1])

    @QtCore.Slot(int)
    @QtCore.Slot(float)
    def update_value(self, new_value):

        # stash numerical value
        try:
            new_value = np.clip(new_value, self.abs_range[0], self.abs_range[1])
        except TypeError:
            # if given None, can't clip it.
            # just return
            # TODO: Should this raise an alarm?
            return
        self.value = new_value
        self.check_alarm()

    @QtCore.Slot(tuple)
    def update_limits(self, new_limits):
        self.range_slider.setLow(new_limits[0])
        self.range_slider.setHigh(new_limits[1])
        self.update_boxes(new_limits)

    def timed_update(self):
        # format value based on decimals
        if self.value:
            value_str = str(np.round(self.value, self.decimals))
            self.value_label.setText(value_str)

        QtCore.QTimer.singleShot(round(self.update_period*1000), self.timed_update)

    def _limits_changed(self, val):
        # ignore value, just emit changes and check alarm
        self.check_alarm()
        self.limits_changed.emit((self.min_safe.value(), self.max_safe.value()))

    @property
    def alarm_state(self):
        return self._alarm

    @alarm_state.setter
    def alarm_state(self, alarm):
        if alarm == True:
            self.value_label.setStyleSheet(styles.DISPLAY_VALUE_ALARM)
            self.name_label.setStyleSheet(styles.DISPLAY_NAME_ALARM)
            self.units_label.setStyleSheet(styles.DISPLAY_UNITS_ALARM)
            self._alarm = True
        elif alarm == False:
            self.value_label.setStyleSheet(styles.DISPLAY_VALUE)
            self.name_label.setStyleSheet(styles.DISPLAY_NAME)
            self.units_label.setStyleSheet(styles.DISPLAY_UNITS)
            self._alarm = False

    @QtCore.Slot(bool)
    def set_alarm(self, alarm):
        """
        Simple wrapper to set alarm state from a qt signal

        Args:
            alarm (bool): Whether to set as alarm state or not
        """
        self.alarm_state = alarm

    def toggle_alarm(self):

        if self.alarm_state == False:
            self.alarm_state = True
        else:
            self.alarm_state = False

    def check_alarm(self, value=None):
        if value is None:
            value = self.value

        if value:
            if (value > self.max_safe.value()) or (value < self.min_safe.value()):
                self.alarm.emit((self.enum_name, value, time.time()))

                if self.alarm == False:
                    self.toggle_alarm_state(True)
            else:
                if self.alarm == True:
                    self.toggle_alarm_state(False)


from .pigpio_mocks import patch_pigpio_base, patch_pigpio_i2c, patch_pigpio_gpio, mock_i2c_hardware, patch_bad_socket
from .pigpio_mocks import MockHardwareDevice
from vent.io.devices import ADS1115, ADS1015, SPIDevice, PigpioConnection, IODeviceBase, I2CDevice

from secrets import token_bytes

import pigpio
import pytest
import random


def test_mock_pigpio_base(patch_pigpio_base):
    """__________________________________________________________________________________________________________TEST #1
    Tests that the base fixture of pigpio mocks is indeed enough to instantiate pigpio.pi()
    """
    pig = PigpioConnection()
    assert isinstance(pig, pigpio.pi)
    assert pig.connected
    pig.stop()
    assert not pig.connected
    """__________________________________________________________________________________________________________
    """


def test_pigpio_connection_exception(patch_pigpio_base, patch_bad_socket):
    """__________________________________________________________________________________________________________TEST #2
    Tests to make sure an exception is thrown if, upon init, a PigpioConnection finds it is not connected.
    """
    with pytest.raises(RuntimeError):
        PigpioConnection()
    with pytest.raises(RuntimeError):
        IODeviceBase()


def test_io_device_base_no_handles_to_close(patch_pigpio_base, monkeypatch):
    """__________________________________________________________________________________________________________TEST #3
    Tests that IODeviceBase._close() returns cleanly when it has no handles
    """
    device = IODeviceBase()
    device._close()


@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(128)])
def test_mock_pigpio_i2c(patch_pigpio_i2c, mock_i2c_hardware, seed):
    """__________________________________________________________________________________________________________TEST #4
    Tests the functionality of the mock pigpio i2c device interface. More specifically, tests that:
    - mock hardware devices are initialized correctly
    - mock handles, i2c_open(), and i2c_close() can be used to add, interact with, and remove mock hardware devices
        - correct functioning of i2c_open is implied by the read/write tests
        - the correct functioning of i2c_close() is ascertained from the assertion that pig.mock_i2c[i2c_bus] is empty
    - mock i2c read and write functions work as intended:
        - This is three tests in one. For each register of each mock hardware device:
            1) Read the register with i2c_read and store the result in results['init']. We expect it to be the value
            from the correct register of the matching mock_device (expected['init'])
            2) Write two random bytes to the register with i2c_write. Look at the actual contents of the target register
                and put them in results['write']. We expect them to be the same as the two random bytes we generated.
            3) Read the register with i2c_read again and put the results in expected['read']. These should also match
                the two random bytes from the 'write' test.
            -> Assert that the results match what we expect
    - if there is only one register on the device, test read_device() and write_device() instead of read_register()
        and write_register() -> this matches how such a device would be interacted with in practice.
    """
    random.seed(seed)
    n_devices = random.randint(1, 20)
    address_pool = random.sample(range(128), k=n_devices)
    pig = PigpioConnection()
    mocks = []
    for i in range(n_devices):
        mock = mock_i2c_hardware(i2c_address=address_pool.pop())
        pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
        mock['handle'] = pig.i2c_open(mock['i2c_bus'], mock['i2c_address'])
        mocks.append(mock)
    for mock_device in mocks:
        handle = mock_device['handle']
        bus = mock_device['i2c_bus']
        address = mock_device['i2c_address']
        results = {'init': [], 'write': [], 'read': []}
        expected = {'init': [], 'write': [], 'read': []}
        for reg in range(len(mock_device['values'])):
            expected['init'] = mock_device['values'][reg]
            expected['write'] = token_bytes(2)
            expected['read'] = expected['write']
            if len(mock_device['values']) == 1:
                results['init'] = pig.i2c_read_device(handle, 2)[1]
                pig.i2c_write_device(handle, expected['write'])
                results['write'] = pig.mock_i2c[bus][address].registers[0][-1]
                results['read'] = pig.i2c_read_device(handle, 2)[1]
            else:
                results['init'] = pig.i2c_read_i2c_block_data(handle, reg, count=2)[1]
                pig.i2c_write_i2c_block_data(handle, reg, expected['write'])
                results['write'] = pig.mock_i2c[bus][address].registers[reg][-1]
                results['read'] = pig.i2c_read_i2c_block_data(handle, reg, count=2)[1]
            assert results == expected
        pig.i2c_close(handle)
    assert not pig.mock_i2c[0]
    assert not pig.mock_i2c[1]
    assert not pig.mock_i2c['spi']
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(128)])
def test_i2c_device(patch_pigpio_i2c, mock_i2c_hardware, seed):
    """__________________________________________________________________________________________________________TEST #5
    Tests the various basic I2CDevice methods, open, close, read,  write, etc.
        - Note: i2c_device.read_device() returns raw bytes (assumed to have big-endian ordering.) read_register, on the
        other hand, returns the integer representation of those bytes. write_device and write_register take an integer
        argument which is the written to the device register as a big-endian two's complement.
        - The RPi is native little endian, so to simulate/test the byte-swapping that occurs over I2C we generate a
        random 16-bit integer (`data`), write it to the register, and read it back.
        - If the int->byte-> int conversions in i2c_device are working correctly, we should expect one of two possible
         values to be returned (depending on whether read_device() or read_register() is called)
            - read_device() should return the big-endian two's complement of `data`
            - read_register() should just return `data`
    """
    results = []
    random.seed(seed)
    mock = mock_i2c_hardware()
    n_registers = len(mock['values'])
    data = [random.getrandbits(15) for _ in range(n_registers)]
    signed = random.getrandbits(1)
    expected = [data[0].to_bytes(2, 'big', signed=signed)] if n_registers == 1 else data
    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    i2c_device = I2CDevice(mock['i2c_address'], mock['i2c_bus'], pig=pig)
    assert i2c_device.pigpiod_ok
    if n_registers == 1:
        i2c_device.write_device(data[0], signed=signed)
        results.append(i2c_device.read_device()[1])
    else:
        for register in range(n_registers):
            i2c_device.write_register(register, data[register], signed=signed)
            results.append(i2c_device.read_register(register, signed=signed))
    i2c_device._close()
    assert not i2c_device._pig.mock_i2c[mock['i2c_bus']]
    assert results == expected
    """__________________________________________________________________________________________________________
    """


def test_register_arg_exception():
    """__________________________________________________________________________________________________________TEST #6
        Tests that I2CDevice.Register throws an exception if it tries to init with a mismatched number of arguments
    """
    with pytest.raises(ValueError):
        I2CDevice.Register(fields=('one', 'two', 'three'), values=((0, 1, 3), (4, 5)))


def test_value_field_pack_unknown_value_exception():
    """__________________________________________________________________________________________________________TEST #7
        Tests that I2CDevice.Register throws an exception if it tries to init with a mismatched number of arguments
    """
    reg = I2CDevice.Register(fields=('one', 'two', 'three'), values=((0, 1, 3), (4, 5), ('hi', 'mark')))
    with pytest.raises(ValueError):
        reg.three.pack('oh')


def test_value_field_insert_unknown_value_exception():
    """__________________________________________________________________________________________________________TEST #8
        Tests that I2CDevice.Register throws an exception if it tries to init with a mismatched number of arguments
    """
    reg = I2CDevice.Register(fields=('one', 'two', 'three'), values=((0, 1, 3), (4, 5), ('hi', 'mark')))
    with pytest.raises(ValueError):
        reg.three.insert(0x0, 'oh')


def test_spi_device(patch_pigpio_i2c):
    """__________________________________________________________________________________________________________TEST #9
        Tests that an SPI device can be created without issue. That's about all you can do with one of these anyway.
    """
    channel = random.randint(1, 20)
    device = SPIDevice(channel=channel, baudrate=100)
    mock_device = MockHardwareDevice(token_bytes(2))
    device.pig.add_mock_hardware(mock_device, i2c_address=channel, i2c_bus='spi')
    assert device._handle >= 0
    device._close()


@pytest.mark.parametrize("ads1x15", [ADS1115, ADS1015])
@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(128)])
def test_read_conversion(patch_pigpio_i2c, mock_i2c_hardware, ads1x15, seed):
    """_________________________________________________________________________________________________________TEST #10
    Tests that the proper cfg is generated and written to the config register given kwargs, and tests that the
        conversion register is properly read & converted to signed int
    """
    random.seed(seed)
    kwargs = {
        "MUX": random.choice(ads1x15._CONFIG_VALUES[1]),
        "PGA": random.choice(ads1x15._CONFIG_VALUES[2]),
        "MODE": random.choice(ads1x15._CONFIG_VALUES[3]),
        "DR": random.choice(ads1x15._CONFIG_VALUES[4])}
    conversion_bytes = token_bytes(2)
    expected_val = int.from_bytes(conversion_bytes, 'big', signed=True) * kwargs['PGA'] / 32767
    mock = mock_i2c_hardware(
        i2c_bus=1,
        i2c_address=ads1x15._DEFAULT_ADDRESS,
        n_registers=4,
        reg_values=[
            conversion_bytes,
            b'\x85\x83'
        ]
    )
    pig = PigpioConnection()
    pig.add_mock_hardware(mock['device'], mock['i2c_address'], mock['i2c_bus'])
    ads = ads1x15(pig=pig)

    expected_default_config = ads.config.unpack(0xC3E3)
    expected_config = ads.config.unpack(ads.config.pack(0xC3E3, **kwargs))
    assert ads.print_config() == expected_default_config
    result = ads.read_conversion(**kwargs)
    assert ads.print_config() == expected_config
    assert result == expected_val
    """__________________________________________________________________________________________________________
    """



from .pigpio_mocks import patch_pigpio_base
from vent.io import Hal
from random import randint, getrandbits


def test_functions(patch_pigpio_base):
    """__________________________________________________________________________________________________________TEST #1
        Using simulated sensors, tests that hal initializes correctly, that sensors can be read, and that all valves
            can be open, closed, and/or controlled using the setpoint properties of Hal.
            - Checks to make sure that valves report as being open if they are not fully closed (i.e. if they have a
                nonzero setpoint)
    """
    hal = Hal(config_file='vent/io/config/sim-devices.ini')
    for _ in range(randint(1000, 10000)):
        assert hal._pressure_sensor.low <= hal.pressure <= hal._pressure_sensor.high
        assert hal._aux_pressure_sensor.low <= hal.aux_pressure <= hal._aux_pressure_sensor.high
        assert hal._flow_sensor_in.low <= hal.flow_in <= hal._flow_sensor_in.high
        assert hal._flow_sensor_ex.low <= hal.flow_ex <= hal._flow_sensor_ex.high
        setpoint_in = randint(0, 100)
        hal.setpoint_in = setpoint_in
        assert hal.setpoint_in == setpoint_in
        if hal.setpoint_in == 0:
            assert not hal._inlet_valve.is_open
        else:
            assert hal._inlet_valve.is_open
        setpoint_ex = getrandbits(1)
        hal.setpoint_ex = setpoint_ex
        assert hal.setpoint_ex == setpoint_ex
        if hal.setpoint_ex == 0:
            assert not hal._expiratory_valve.is_open
        else:
            assert hal._expiratory_valve.is_open



""" Module for interacting with physical and/or simulated devices installed on the ventilator.

"""

from importlib import import_module
from ast import literal_eval
from .devices.sensors import Sensor
from .devices.base import PigpioConnection

import vent.io.devices.valves as valves
import configparser


class Hal:
    """ Hardware Abstraction Layer for ventilator hardware.
    Defines a common API for interacting with the sensors & actuators on the ventilator. The types of devices installed
    on the ventilator (real or simulated) are specified in a configuration file.
    """

    def __init__(self, config_file='vent/io/config/devices.ini'):
        """ Initializes HAL from config_file.
            For each section in config_file, imports the class <type> from module <module>, and sets attribute
            self.<section> = <type>(**opts), where opts is a dict containing all of the options in <section> that are
            not <type> or <section>. For example, upon encountering the following entry in config_file.ini:

                [adc]
                type   = ADS1115
                module = devices
                i2c_address = 0x48
                i2c_bus = 1

            The Hal will:
                1) Import vent.io.devices.ADS1115 (or ADS1015) as a local variable:
                        class_ = getattr(import_module('.devices', 'vent.io'), 'ADS1115')

                2) Instantiate an ADS1115 object with the arguments defined in config_file and set it as an attribute:
                        self._adc = class_(pig=self.-pig,address=0x48,i2c_bus=1)

            Note: RawConfigParser.optionxform() is overloaded here s.t. options are case sensitive (they are by default
            case insensitive). This is necessary due to the kwarg MUX which is so named for consistency with the config
            registry documentation in the ADS1115 datasheet. For example, A P4vMini pressure_sensor on pin A0 (MUX=0)
            of the ADC is passed arguments like:

            analog_sensor = AnalogSensor(
                pig=self._pig,
                adc=self._adc,
                MUX=0,
                offset_voltage=0.25,
                output_span = 4.0,
                conversion_factor=2.54*20
            )

            Note: ast.literal_eval(opt) interprets integers, 0xFF, (a, b) etc. correctly. It does not interpret strings
            correctly, nor does it know 'adc' -> self._adc; therefore, these special cases are explicitly handled.
        Args:
            config_file (str): Path to the configuration file containing the definitions of specific components on the
                ventilator machine. (e.g., config_file = "vent/io/config/devices.ini")
        """
        self._setpoint_in = 0.0  # setpoint for inspiratory side
        self._setpoint_ex = 0.0  # setpoint for expiratory side
        self._adc = object
        self._inlet_valve = object
        self._control_valve = object
        self._expiratory_valve = object
        self._pressure_sensor = object
        self._aux_pressure_sensor = object
        self._flow_sensor_in = object
        self._flow_sensor_ex = object
        self._pig = PigpioConnection(show_errors=False)
        self.config = configparser.RawConfigParser()
        self.config.optionxform = lambda option: option
        self.config.read(config_file)
        for section in self.config.sections():
            sdict = dict(self.config[section])
            class_ = getattr(import_module('.' + sdict['module'], 'vent.io'), sdict['type'])
            opts = {key: sdict[key] for key in sdict.keys() - ('module', 'type',)}
            for key in opts.keys():
                if key == 'adc':
                    opts[key] = self._adc
                else:
                    opts[key] = literal_eval(opts[key])
            print("  [ {device_name:^19} ]  opts: {device_options}".format(
                device_name=section,
                device_options=opts
            ))  # debug
            setattr(self, '_' + section, class_(pig=self._pig, **opts))

    # TODO: Need exception handling whenever inlet valve is opened

    @property
    def pressure(self) -> float:
        """ Returns the pressure from the primary pressure sensor.
        """
        return self._pressure_sensor.get()

    @property
    def aux_pressure(self) -> float:
        """ Returns the pressure from the auxiliary pressure sensor, if so equipped.
        If a secondary pressure sensor is not defined, raises a RuntimeWarning.
        """
        if isinstance(self._aux_pressure_sensor, Sensor):
            return self._aux_pressure_sensor.get()
        else:
            raise RuntimeWarning('Secondary pressure sensor not instantiated. Check your "devices.ini" file.')

    @property
    def flow_in(self) -> float:
        """ The measured flow rate inspiratory side."""
        return self._flow_sensor_in.get()

    @property
    def flow_ex(self) -> float:
        """ The measured flow rate expiratory side."""
        return self._flow_sensor_ex.get()

    @property
    def setpoint_in(self) -> float:
        """ The currently requested flow for the inspiratory proportional control valve as a proportion of maximum."""
        return self._setpoint_in

    @setpoint_in.setter
    def setpoint_in(self, value: float):
        """ Sets the openness of the inspiratory valve to the requested value.

        Args:
            value: Requested flow, as a proportion of maximum. Must be in [0, 1].
        """
        if not 0 <= value <= 100:
            raise ValueError('setpoint must be a number between 0 and 100')
        if value > 0 and not self._inlet_valve.is_open:
            self._inlet_valve.open()
        elif value == 0 and self._inlet_valve.is_open:
            self._inlet_valve.close()
        self._control_valve.setpoint = value
        self._setpoint_in = value

    @property
    def setpoint_ex(self) -> float:
        """ The currently requested flow on the expiratory side as a proportion of the maximum."""
        return self._setpoint_ex

    @setpoint_ex.setter
    def setpoint_ex(self, value):
        """ Sets the openness of the expiratory valve to the requested value.

        Args:
            value (float): Requested flow, as a proportion of maximum. Must be either 0 or 1 for OnOffValve, and between
                0 and 1 for a (proportional) control valve.
        """
        if (
                isinstance(self._expiratory_valve, valves.OnOffValve) or
                isinstance(self._expiratory_valve, valves.SimOnOffValve)
        ):
            if value not in (0, 1):
                raise ValueError('setpoint must be either 0 or 1 for an On/Off expiratory valve')
            elif value == 1:
                self._expiratory_valve.open()
            else:
                self._expiratory_valve.close()
        elif (
                isinstance(self._expiratory_valve, valves.PWMControlValve) or
                isinstance(self._expiratory_valve, valves.SimControlValve)
        ):
            if not 0 <= value <= 100:
                raise ValueError('setpoint must be between 0 and 100 for an expiratory control valve')
            else:
                self._expiratory_valve.setpoint = value
        self._setpoint_ex = value



import time
import importlib

from itertools import count
from vent.alarm import AlarmType, AlarmSeverity

class Alarm:
    """
    Class used by the program to control and coordinate alarms.

    Parameterized by a :class:`Alarm_Rule` and managed by :class:`Alarm_Manager`
    """


    id_counter = count()
    """
    :class:`itertools.count`: used to generate unique IDs for each alarm
    """

    def __init__(self,
                 alarm_type: AlarmType,
                 severity: AlarmSeverity,
                 start_time: float = None,
                 latch = True,
                 persistent = True,
                 value=None,
                 message=None,
                 managed = False):
        """

        Args:
            alarm_type :
            severity:
            start_time:
            value (int, float): optional - numerical value that generated the alarm
            message (str): optional - override default text generated by :class:`~vent.gui.alarm_manager.AlarmManager`
            managed (bool): if created by the alarm_manager, don't register
        """


        self.id = next(Alarm.id_counter)

        assert isinstance(severity, AlarmSeverity)
        self._severity = severity

        assert isinstance(alarm_type, AlarmType)
        self._alarm_type = alarm_type

        if start_time is None:
            self.start_time = time.time()
        else:
            assert isinstance(start_time, float)
            self.start_time = start_time

        self.active = True
        if self.severity.value == AlarmSeverity.OFF.value:
            self.active = False

        self.alarm_end_time = None
        self.value = value
        self.message = message
        self.latch = latch
        self.persistent = persistent

        if not managed:
            self.manager.register_alarm(self)

    @property
    def manager(self):
        """
        have ta do it this janky way to avoid circular imports
        """
        try:
            return Alarm_Manager()
        except:
            # import into the module namespace
            manager_module = importlib.import_module('vent.alarm.alarm_manager')
            globals()['Alarm_Manager'] = getattr(manager_module, 'Alarm_Manager')
            return Alarm_Manager()

    @property
    def severity(self):
        # no setter, don't want to be able to change after instantiation
        return self._severity

    @property
    def alarm_type(self):
        return self._alarm_type



    def deactivate(self):
        if not self.active:
            return

        self.alarm_end_time = time.time()
        self.active = False
        # make sure the manager deactivates us.
        # manager checks if this has already been done so doesn't recurse
        self.manager.deactivate_alarm(self)



"""
test objects
* main
* components
* control
* monitor
* plot
* status_bar

test types
* user interaction
* flooding


"""

from copy import copy
import pdb

import pytest
from time import sleep


from pytestqt import qt_compat
from pytestqt.qt_compat import qt_api

# mock before importing


from vent import gui
from vent.gui import styles
from vent.gui import widgets
from vent.common import message, values
from vent.coordinator.coordinator import get_coordinator


from PySide2 import QtCore, QtGui, QtWidgets

import numpy as np

##################################




# turn off gui limiting
gui.limit_gui(False)

n_samples = 100
decimals = 5
global_minmax_range = (0, 100)
global_safe_range = (25, 75)

@pytest.fixture
def generic_minmax():
    """
    Make a (min, max) range of values that has some safe space in between.

    ie. min can be between 0 and 0.25*(minmax_range[1]-minmax_range[0])
    and max can similarly be above the 75% mark of the range
    Returns:

    """
    def _generic_minmax():
        abs_min = np.random.rand() * (global_minmax_range[1]-global_minmax_range[0]) * 0.25 + \
                  global_minmax_range[0]
        abs_max = np.random.rand() * (global_minmax_range[1]-global_minmax_range[0]) * 0.25 + \
                  global_minmax_range[0] + (global_minmax_range[1]-global_minmax_range[0]) * 0.75
        abs_min, abs_max = np.round(abs_min, decimals), np.round(abs_max, decimals)
        return abs_min, abs_max
    return _generic_minmax

@pytest.fixture
def generic_saferange():
    """
    Make a (min, max) range of values that has some safe space in between.

    ie. min can be between 0 and 0.25*(minmax_range[1]-minmax_range[0])
    and max can similarly be above the 75% mark of the range
    Returns:

    """
    def _generic_saferange():
        abs_min = np.random.rand() * (global_safe_range[1]-global_safe_range[0]) * 0.25 + \
                  global_safe_range[0]
        abs_max = np.random.rand() * (global_safe_range[1]-global_safe_range[0]) * 0.25 + \
                  global_safe_range[0] + (global_safe_range[1]-global_safe_range[0]) * 0.75
        abs_min, abs_max = np.round(abs_min, decimals), np.round(abs_max, decimals)
        return abs_min, abs_max
    return _generic_saferange

@pytest.fixture(params=[True, False])
def spawn_gui(qtbot, request):
    assert qt_api.QApplication.instance() is not None

    app = qt_api.QApplication.instance()
    app.setStyle('Fusion')
    app.setStyleSheet(styles.DARK_THEME)
    app = styles.set_dark_palette(app)

    coordinator = get_coordinator(sim_mode=True, single_process=request.param)
    vent_gui = gui.Vent_Gui(coordinator)
    #app, vent_gui = launch_gui(coordinator)
    qtbot.addWidget(vent_gui)
    return app, vent_gui




def test_gui_launch(qtbot, spawn_gui):

    app, vent_gui = spawn_gui
    vent_gui.status_bar.start_button.click()

    # wait for a second to let the simulation spin up and start spitting values
    qtbot.wait(2000)

    assert vent_gui.isVisible()

@pytest.mark.timeout(15)
def test_gui_launch_mp(qtbot):
    assert qt_api.QApplication.instance() is not None

    coordinator = get_coordinator(sim_mode=True, single_process=False)
    coordinator.start()

    vent_gui = gui.Vent_Gui(coordinator)
    qtbot.addWidget(vent_gui)
    vent_gui.status_bar.start_button.click()

    # wait for a second to let the simulation spin up and start spitting values
    qtbot.wait(5000)

    assert vent_gui.isVisible()


################################
# test user interaction

@pytest.mark.parametrize("test_value", [(k, v) for k, v in values.CONTROL.items()])
def test_gui_controls(qtbot, spawn_gui, test_value):
    """
    test setting controls in all the ways available to the GUI

    from the :class:`~vent.gui.widgets.control.Control` widget:

        * :class:`~vent.gui.widgets.components.EditableLabel` - setting label text
        * setting slider value
        * using :meth:`~vent.gui.main.Vent_Gui.set_value`


    Args:
        qtbot:
        spawn_gui:
        test_value:


    """

    app, vent_gui = spawn_gui

    vent_gui.start()
    vent_gui.timer.stop()

    value_name = test_value[0]
    value_params = test_value[1]
    abs_range = value_params.abs_range

    # generate target value
    def gen_test_value():
        test_value = np.random.rand()*(abs_range[1]-abs_range[0]) + abs_range[0]
        test_value = np.round(test_value, value_params.decimals)
        return test_value

    ####
    # test setting controls from control widget
    # from editablelabel

    control_widget = vent_gui.controls[value_name.name]

    for i in range(n_samples):
        test_value = gen_test_value()

        control_widget.value_label.setLabelEditableAction()
        control_widget.value_label.lineEdit.setText(str(test_value))
        control_widget.value_label.returnPressedAction()
        # should call labelUpdatedAction and send to controller

        control_value = vent_gui.coordinator.get_control(value_name)

        assert(control_value.value == test_value)

    # from slider
    # toggle it open
    assert(control_widget.slider_frame.isVisible() == False)
    control_widget.toggle_button.click()
    assert(control_widget.slider_frame.isVisible() == True)

    for i in range(n_samples):
        test_value = gen_test_value()
        control_widget.slider.setValue(test_value)

        control_value = vent_gui.coordinator.get_control(value_name)
        assert(control_value.value == test_value)

    # from set_value
    for i in range(n_samples):
        test_value = gen_test_value()
        vent_gui.set_value(test_value, value_name = value_name)

        control_value = vent_gui.coordinator.get_control(value_name)
        assert(control_value.value == test_value)


@pytest.mark.parametrize("test_value", [(k, v) for k, v in values.SENSOR.items()])
def test_gui_monitor(qtbot, spawn_gui, test_value):


    app, vent_gui = spawn_gui

    vent_gui.start()
    vent_gui.timer.stop()


    value_name = test_value[0]
    value_params = test_value[1]
    abs_range = value_params.abs_range

    # generate target value
    def gen_test_values():
        test_value = np.random.rand(2)*(abs_range[1]-abs_range[0]) + abs_range[0]
        test_value = np.round(test_value, value_params.decimals)
        return np.min(test_value), np.max(test_value)

    monitor_widget = vent_gui.monitor[value_name.name]

    # open the control
    assert(monitor_widget.slider_frame.isVisible() == False)
    monitor_widget.toggle_button.click()
    assert (monitor_widget.slider_frame.isVisible() == True)

    # set handles to abs_min and max so are on absolute right and left sides
    monitor_widget.range_slider.setValue(abs_range)
    assert(monitor_widget.range_slider.low == monitor_widget.range_slider.minimum())
    assert (monitor_widget.range_slider.high == monitor_widget.range_slider.maximum())
    #
    # # move left a quarter of the way to the right
    # widget_size = monitor_widget.range_slider.size()
    #
    # # get low box position
    # low_pos = monitor_widget.range_slider.get_handle_rect(0)
    # click_pos = low_pos.center()
    # move_pos = copy(click_pos)
    # move_pos.setX(move_pos.x() + (widget_size.width()/4))
    #
    # qtbot.mouseMove(monitor_widget.range_slider, click_pos, delay=100)
    # qtbot.mousePress(monitor_widget.range_slider, QtCore.Qt.LeftButton, delay=200)
    # qtbot.mouseMove(monitor_widget.range_slider, move_pos)
    # qtbot.mouseRelease(monitor_widget.range_slider, QtCore.Qt.LeftButton, pos=move_pos, delay=200)


    # set with range_slider
    # for i in range(n_samples):
    #     test_min, test_max = gen_test_values()
    #
    #
    #     with qtbot.waitSignal(monitor_widget.limits_changed, timeout=1000) as blocker:
    #         monitor_widget.range_slider.setLow(test_min)
    #         sleep(0.01)
    #         assert(blocker.args[0][0]==test_min)
    #
    #     with qtbot.waitSignal(monitor_widget.limits_changed, timeout=1000) as blocker:
    #         monitor_widget.range_slider.setHigh(test_max)
    #         sleep(0.01)
    #         assert(blocker.args[0][1]==test_max)









###################################
# Test base components

#####
# doubleslider

def test_doubleslider(qtbot):
    """
    test that the doubleslider accurately represents floats
    """
    doubleslider = widgets.components.DoubleSlider(decimals=decimals)
    doubleslider.show()
    qtbot.addWidget(doubleslider)

    doubleslider.setMinimum(0)
    doubleslider.setMaximum(1)

    for i in np.random.rand(n_samples):
        test_val = np.round(i, decimals)

        # set value inside of signal catcher to test both value and signal
        with qtbot.waitSignal(doubleslider.doubleValueChanged, timeout=1000) as blocker:
            doubleslider.setValue(test_val)

        assert(doubleslider.value() == test_val)
        assert(blocker.args == test_val)


def test_doubleslider_minmax(qtbot, generic_minmax):

    doubleslider = widgets.components.DoubleSlider(decimals=decimals)
    doubleslider.show()
    qtbot.addWidget(doubleslider)

    multiplier = 100
    for i in range(n_samples):
        min, max = generic_minmax()

        doubleslider.setMinimum(min)
        doubleslider.setMaximum(max)

        # test that values were set correctly
        assert(doubleslider.minimum() == min)
        assert(doubleslider.maximum() == max)

        # test below min and above max
        test_min = min - np.random.rand()*multiplier
        test_max = max + np.random.rand()*multiplier

        doubleslider.setValue(test_min)
        assert(doubleslider.value() == doubleslider.minimum())
        doubleslider.setValue(test_max)
        assert(doubleslider.value() == doubleslider.maximum())

#################
# RangeSlider

def test_rangeslider(qtbot, generic_minmax, generic_saferange):
    abs_range = generic_minmax()
    safe_range = generic_saferange()
    print(safe_range)
    orientation = QtCore.Qt.Orientation.Horizontal


    rangeslider = widgets.components.RangeSlider(
        abs_range,
        safe_range,
        decimals=decimals,
        orientation=orientation)
    rangeslider.show()
    qtbot.addWidget(rangeslider)

    # test ranges & values
    assert(rangeslider.minimum() == abs_range[0])
    assert(rangeslider.maximum() == abs_range[1])
    assert(rangeslider.low == safe_range[0])
    assert(rangeslider.high == safe_range[1])

    for i in range(n_samples):
        min, max = generic_saferange()
        print(min, max)

        with qtbot.waitSignal(rangeslider.valueChanged, timeout=1000) as blocker:
            rangeslider.setHigh(max)
            assert(rangeslider.high == max)

            rangeslider.setLow(min)
            assert(rangeslider.low == min)

def test_rangeslider_minmax(qtbot, generic_minmax, generic_saferange):
    abs_range = generic_minmax()
    safe_range = generic_saferange()
    decimals = 5
    orientation = QtCore.Qt.Orientation.Horizontal

    rangeslider = widgets.components.RangeSlider(
        abs_range,
        safe_range,
        decimals=decimals,
        orientation=orientation)
    rangeslider.show()
    qtbot.addWidget(rangeslider)

    for i in range(n_samples):
        abs_min, abs_max = generic_minmax()
        safe_min, safe_max = generic_saferange()

        #pdb.set_trace()
        # Set min and max and test they were set correctly
        rangeslider.setMinimum(abs_min)
        assert(rangeslider.minimum() == abs_min)

        rangeslider.setMaximum(abs_max)
        assert(rangeslider.maximum() == abs_max)

        # set low and high and test they were set correctly
        rangeslider.setHigh(safe_max)
        assert(rangeslider.high == safe_max)

        rangeslider.setLow(safe_min)
        assert(rangeslider.low == safe_min)

        # try to set low and high outside of max
        rangeslider.setHigh(abs_max + 1)
        assert(rangeslider.high == rangeslider.maximum())

        rangeslider.setLow(abs_min - 1)
        assert(rangeslider.low == rangeslider.minimum())

        # try to set low higher and high and vice versa
        midpoint = np.round(np.mean([abs_min, abs_max]), decimals)
        highpoint = np.round(midpoint-1, decimals)
        lowpoint = np.round(midpoint-1-(10**-decimals), decimals)

        rangeslider.setLow(midpoint)
        rangeslider.setHigh(highpoint)

        assert(rangeslider.high == highpoint)
        assert(rangeslider.low == lowpoint)

        highpoint = np.round(midpoint+1+(10**-decimals), decimals)
        lowpoint = np.round(midpoint + 1, decimals)

        rangeslider.setHigh(midpoint)
        rangeslider.setLow(lowpoint)

        assert(rangeslider.low == lowpoint)
        assert(rangeslider.high == highpoint)




from setuptools import setup, find_packages
import os
import subprocess

depend_links = []

# detect if on raspberry pi, and
# set location to wheel if we are
IS_RASPI = False
ret = subprocess.call(['grep', '-q', 'BCM', '/proc/cpuinfo'])
if ret == 0:
    IS_RASPI = True
    os.system("sudo +x INSTALL")
    os.system("sudo ./INSTALL")

    # keeping this around for proper packaging later
    # get wheel name
    #external_files = os.listdir(os.path.join(os.getcwd(), 'external'))
    #pyside_wheel = [whl for whl in external_files if whl.endswith('.whl') and whl.startswith("PySide2")][0]
    #depend_links.append(os.path.join(os.getcwd(), 'external', pyside_wheel))


setup(
    name="pvp",
    author="vent team",
    author_email="vent@vents.com",
    description="some description of how we made a ventilator",
    keywords="vents ventilators etc",
    url="https://ventilator.readthedocs.io",
    version="0.0.2",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'PySide2==5.11.*',
        'pyqtgraph>=0.11.0rc0',
        'pytest-qt',
        'pytest-timeout',
        'pigpio'
    ],
    dependency_links=depend_links,
    python_requires='==3.7.*'
)



from enum import Enum, auto

class AlarmType(Enum):
    LOW_PRESSURE  = auto()  # low airway pressure alarm
    HIGH_PRESSURE = auto()  # high airway pressure alarm
    LOW_VTE       = auto()  # low VTE
    HIGH_VTE      = auto()
    LOW_PEEP      = auto()
    HIGH_PEEP     = auto()
    LOW_O2        = auto()
    HIGH_O2       = auto()
    OBSTRUCTION   = auto()
    LEAK          = auto()


class AlarmSeverity(Enum):
    HIGH = 3
    MEDIUM = 2
    LOW = 1
    OFF = 0

from vent.alarm import condition
from vent.alarm import AlarmType, AlarmSeverity
from vent.alarm.rule import Alarm_Rule
from vent.alarm.alarm import Alarm
from vent.alarm.alarm_manager import Alarm_Manager
from vent.common.values import ValueName, VALUES
from collections import OrderedDict as odict


ALARM_RULES = odict({
    AlarmType.LOW_PRESSURE: Alarm_Rule(
        name = AlarmType.LOW_PRESSURE,
        latch = False,
        persistent = False,
        conditions = (
            (
            AlarmSeverity.LOW,
                condition.ValueCondition(
                    value_name=ValueName.PRESSURE,
                    limit=VALUES[ValueName.PIP]['safe_range'][0],
                    mode='min',
                    depends={
                        'value_name': ValueName.PIP,
                        'value_attr': 'value',
                        'condition_attr': 'limit',
                        'transform': lambda x : x-(x*0.10)
                    }
                )
            ),
            (
            AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.PRESSURE,
                    limit=VALUES[ValueName.PIP]['safe_range'][0]- \
                          VALUES[ValueName.PIP]['safe_range'][0]*0.15,
                    mode='min'
                ) + \
                condition.CycleAlarmSeverityCondition(
                    alarm_type = AlarmType.LOW_PRESSURE,
                    severity   = AlarmSeverity.LOW,
                    n_cycles = 2
                )
            )
        )
    ),
    AlarmType.HIGH_PRESSURE: Alarm_Rule(
        name = AlarmType.HIGH_PRESSURE,
        latch = True,
        persistent = True,
        conditions = (
            (
            AlarmSeverity.HIGH,
                condition.ValueCondition(
                    value_name=ValueName.PRESSURE,
                    limit=VALUES[ValueName.PIP]['safe_range'][1],
                    mode='max'
                )
            ),
        )
    ),
    AlarmType.LOW_VTE: Alarm_Rule(
        name = AlarmType.LOW_VTE,
        latch = False,
        persistent = False,
        conditions = (
            (
            AlarmSeverity.LOW,
                condition.ValueCondition(
                    value_name=ValueName.VTE,
                    limit=VALUES[ValueName.VTE]['safe_range'][0],
                    mode='min'
                )
            ),
            (
            AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.VTE,
                    limit=VALUES[ValueName.VTE]['safe_range'][0]- \
                          VALUES[ValueName.VTE]['safe_range'][0]*0.15,
                    mode='min'
                ) + \
                condition.CycleAlarmSeverityCondition(
                    alarm_type = AlarmType.LOW_VTE,
                    severity   = AlarmSeverity.LOW,
                    n_cycles = 2
                )
            )
        )
    ),
    AlarmType.HIGH_VTE: Alarm_Rule(
        name = AlarmType.HIGH_VTE,
        latch = False,
        persistent = False,
        conditions = (
            (
            AlarmSeverity.LOW,
                condition.ValueCondition(
                    value_name=ValueName.VTE,
                    limit=VALUES[ValueName.VTE]['safe_range'][1],
                    mode='max'
                )
            ),
            (
            AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.VTE,
                    limit=VALUES[ValueName.VTE]['safe_range'][1]+ \
                          VALUES[ValueName.VTE]['safe_range'][1]*0.15,
                    mode='max'
                ) + \
                condition.CycleAlarmSeverityCondition(
                    alarm_type = AlarmType.HIGH_VTE,
                    severity   = AlarmSeverity.LOW,
                    n_cycles = 2
                )
            )
        )
    ),
    AlarmType.LOW_PEEP: Alarm_Rule(
        name = AlarmType.LOW_PEEP,
        latch = False,
        persistent = False,
        conditions = (
            (
            AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.PEEP,
                    limit=VALUES[ValueName.PEEP]['safe_range'][0],
                    mode='min'
                )
            ),
        )
    ),
    AlarmType.HIGH_PEEP: Alarm_Rule(
        name = AlarmType.HIGH_PEEP,
        latch = False,
        persistent = False,
        conditions = (
            (
            AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.PEEP,
                    limit=VALUES[ValueName.PEEP]['safe_range'][1],
                    mode='max'
                )
            ),
        )
    ),

    AlarmType.LOW_O2: Alarm_Rule(
        name=AlarmType.LOW_O2,
        latch=False,
        persistent=False,
        conditions=(
            (
                AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.FIO2,
                    limit=VALUES[ValueName.FIO2]['safe_range'][0],
                    mode='min'
                )
            ),
        )
    ),
    AlarmType.HIGH_O2: Alarm_Rule(
        name=AlarmType.HIGH_O2,
        latch=False,
        persistent=False,
        conditions=(
            (
                AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.FIO2,
                    limit=VALUES[ValueName.FIO2]['safe_range'][1],
                    mode='max'
                )
            ),
        )
    ),
})




"""
Class to declare alarm rules
"""

from vent.alarm import AlarmType, AlarmSeverity

class Alarm_Rule(object):
    """
    * name of rule
    * conditions: ((alarm_type, (condition_1, condition_2)), ...)
    * persistent (bool): if True, alarm will not be visually dismissed until alarm conditions are no longer true
    * latch (bool): if True, alarm severity cannot be decremented until user manually dismisses

    * silencing/overriding rules
    """

    def __init__(self, name: AlarmType, conditions, latch=True, persistent=True, technical=False):
        super(Alarm_Rule, self).__init__()

        self.name = name
        self.conditions = conditions
        self.technical = technical
        self.latch = latch
        self.persistent = persistent

        self._severity = AlarmSeverity.OFF

    def check(self, sensor_values):
        """
        Check all of our :attr:`.conditions` .

        Args:
            sensor_values:

        Returns:

        """
        active_severity = AlarmSeverity.OFF
        for severity, condition in self.conditions:
            active = condition.check(sensor_values)
            if active:
                if severity.value > active_severity.value:
                    active_severity = severity

        self._severity = active_severity
        return active_severity

    @property
    def severity(self):
        """
        Last Alarm Severity from ``.check()``
        Returns:
            :class:`.AlarmSeverity`

        """
        return self._severity


    def reset(self):
        for _, condition in self.conditions:
            condition.reset()



import os

_CONFIG = {}

_DIRECTORIES = {}
"""
Directories to ensure are created and added to prefs.

    * ``VENT_DIR``: ~/vent - base directory for user storage
    * ``LOG_DIR``: ~/vent/logs - for storage of event and alarm logs
    * ``DATA_DIR``: ~/vent/data - for storage of waveform data
"""
_DIRECTORIES['VENT_DIR'] = os.path.join(os.path.expanduser('~'), 'vent')
_DIRECTORIES['LOG_DIR'] = os.path.join(_DIRECTORIES['VENT_DIR'], 'logs')
_DIRECTORIES['DATA_DIR'] = os.path.join(_DIRECTORIES['VENT_DIR'], 'logs')

def set_pref(key: str, val):
    globals()['_CONFIG'][key] = val

def get_pref(key: str = None):
    """
    Get global configuration value

    Args:
        key (str, None): get configuration value with specific ``key`` .
            if ``None`` , return all config values.
    """
    if key is None:
        return globals()['_CONFIG']
    else:
        try:
            return globals()['_CONFIG'][key]
        except KeyError:
            return None


def make_dirs():
    """
    ensures _DIRECTORIES are created and added to prefs.
    """
    global _DIRECTORIES
    # create directories if they don't exist already
    for dir_name, make_dir in _DIRECTORIES.items():
        if not os.path.exists(make_dir):
            os.mkdir(make_dir)

        set_pref(dir_name, make_dir)

def init():
    """
    Initialize prefs. Called in ``vent.__init__.py`` to ensure prefs are initialized before anything else.
    """

    # add more functions as needed, but probably bad to hardcode default prefs here.
    # pull them up top like _DIRECTORIES

    make_dirs()


import logging
import time
import datetime

import numpy as np
from PySide2 import QtWidgets, QtCore

from vent.gui import styles, mono_font
from vent.gui import get_gui_instance
from vent.alarm import AlarmSeverity, Alarm


class Status_Bar(QtWidgets.QWidget):
    """
    * Start/stop button
    * Status indicator - a clock that increments with heartbeats,
        or some other visual indicator that things are alright
    * Status bar - most recent alarm or notification w/ means of clearing
    * Override to give 100% oxygen and silence all alarms

    """

    def __init__(self):
        super(Status_Bar, self).__init__()

        self.init_ui()

    def init_ui(self):

        self.layout = QtWidgets.QHBoxLayout()


        self.status_message = Message_Display()
        self.layout.addWidget(self.status_message)
        self.layout.setContentsMargins(5,5,5,5)

        self.heartbeat = HeartBeat()
        self.heartbeat.start_timer()
        self.layout.addWidget(self.heartbeat)

        self.start_button = QtWidgets.QPushButton('start!!!')
        self.layout.addWidget(self.start_button)


        self.setLayout(self.layout)

        style = self.style()
        size = style.pixelMetric(QtWidgets.QStyle.PM_MessageBoxIconSize, None, self)

        # self.setMaximumHeight(size*1.5)
        self.setContentsMargins(0,0,0,0)
        #
        # self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                    QtWidgets.QSizePolicy.Expanding)


class Message_Display(QtWidgets.QFrame):

    message_cleared = QtCore.Signal()
    level_changed = QtCore.Signal()

    def __init__(self):
        super(Message_Display, self).__init__()

        self.icons = {}
        self.alarms = {}
        self.current_alarm = None
        self._alarm_level = AlarmSeverity.OFF
        self.setContentsMargins(0,0,0,0)

        self.make_icons()
        self.init_ui()

    def make_icons(self):

        style = self.style()
        size = style.pixelMetric(QtWidgets.QStyle.PM_MessageBoxIconSize, None, self)

        alarm_icon = style.standardIcon(QtWidgets.QStyle.SP_MessageBoxCritical, None, self)
        alarm_icon = alarm_icon.pixmap(size,size)

        warning_icon = style.standardIcon(QtWidgets.QStyle.SP_MessageBoxWarning, None, self)
        warning_icon = warning_icon.pixmap(size,size)

        normal_icon = style.standardIcon(QtWidgets.QStyle.SP_MessageBoxInformation, None, self)
        normal_icon = normal_icon.pixmap(size,size)

        self.icons[AlarmSeverity.LOW] = normal_icon
        self.icons[AlarmSeverity.MEDIUM] = warning_icon
        self.icons[AlarmSeverity.HIGH] = alarm_icon



    def init_ui(self):

        self.layout = QtWidgets.QHBoxLayout()

        self.message = QtWidgets.QLabel('')
        self.message.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.message.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)

        self.icon = QtWidgets.QLabel()
        #self.icon.setEnabled(False)

        style = self.style()
        size = style.pixelMetric(QtWidgets.QStyle.PM_MessageBoxIconSize, None, self)

        #self.icon.setPixmap()
        self.icon.setFixedHeight(size)
        self.icon.setFixedWidth(size)


        self.clear_button = QtWidgets.QPushButton('Clear Message')
        self.clear_button.setVisible(False)
        self.clear_button.clicked.connect(self.clear_message)
        #clear_icon = QtGui.QIcon.fromTheme('window-close')
        #self.clear_button.setIcon(clear_icon)

        self.layout.addWidget(self.message, 5)
        self.layout.addWidget(self.icon, 1)
        self.layout.addWidget(self.clear_button,1)
        self.setLayout(self.layout)
        self.setFrameStyle(QtWidgets.QFrame.StyledPanel | QtWidgets.QFrame.Raised)

    def draw_state(self, state=None):

        if state == AlarmSeverity.LOW:
            self.setStyleSheet(styles.STATUS_NORMAL)
            self.icon.setPixmap(self.icons[AlarmSeverity.LOW])
            self.clear_button.setVisible(True)
        elif state == AlarmSeverity.MEDIUM:
            self.setStyleSheet(styles.STATUS_WARN)
            self.icon.setPixmap(self.icons[AlarmSeverity.MEDIUM])
            self.clear_button.setVisible(True)
        elif state == AlarmSeverity.HIGH:
            self.setStyleSheet(styles.STATUS_ALARM)
            self.icon.setPixmap(self.icons[AlarmSeverity.HIGH])
            self.clear_button.setVisible(True)
        else:
            self.setStyleSheet(styles.STATUS_NORMAL)
            self.clear_button.setVisible(False)
            self.icon.clear()


    @QtCore.Slot(Alarm)
    def update_message(self, alarm):
        """
        Arguments:
            alarm (:class:`~.message.Alarm`)

        """

        if alarm is None:
            # clear
            self.current_alarm = None
            self.draw_state()
            self.message.setText("")
            return

        self.alarms[alarm.id] = alarm

        if self.current_alarm:
            # see if we are outranked by current message
            if alarm.severity.value >= self.current_alarm.severity.value:
                self.draw_state(alarm.severity)
                self.message.setText(alarm.message)
                self.current_alarm = alarm
                self.alarm_level = alarm.severity
            else:
                return

        else:
            self.draw_state(alarm.severity)
            self.message.setText(alarm.message)
            self.current_alarm = alarm
            self.alarm_level = alarm.severity

        # delete old messages from same value
        self.alarms = {a_key: a_val for a_key, a_val in self.alarms.items() if
                       (a_val.alarm_name != alarm.alarm_name) or
                       (a_val.id == alarm.id)}

    def clear_message(self):
        if not self.current_alarm:
            return

        self.message_cleared.emit(self.current_alarm)
        del self.alarms[self.current_alarm.id]


        # check if we have another message to display
        if len(self.alarms)>0:
            # get message priorities
            paired_priorities = [(alarm.id, alarm.severity.value) for alarm in self.alarms.values()]
            priorities = np.array([msg[1] for msg in paired_priorities])
            # find the max priority
            max_ind = np.argmax(priorities)
            self.current_alarm = None
            new_alarm = self.alarms[paired_priorities[max_ind][0]]
            self.update_message(new_alarm)
            self.alarm_level = new_alarm.severity
        else:
            self.update_message(None)
            self.alarm_level = AlarmSeverity.OFF

    @property
    def alarm_level(self):
        return self._alarm_level

    @alarm_level.setter
    def alarm_level(self, alarm_level):
        if alarm_level != self._alarm_level:
            self.level_changed.emit(alarm_level)
            self._alarm_level = alarm_level


class HeartBeat(QtWidgets.QFrame):

    timeout = QtCore.Signal(bool)
    heartbeat = QtCore.Signal(float)

    def __init__(self, update_interval = 100, timeout_dur = 5000):
        """
        Args:
            update_interval (int): How often to do the heartbeat, in ms
            timeout (int): how long to wait before hearing from control process
        """

        super(HeartBeat, self).__init__()

        self.update_interval = update_interval
        self.start_time = None
        self.timeout_dur = timeout_dur
        self._state = False
        self._last_heartbeat = 0
        self.init_ui()

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self._heartbeat)

        get_gui_instance().gui_closing.connect(self.timer.stop)

    def init_ui(self):

        self.layout = QtWidgets.QGridLayout()

        self.timer_label = QtWidgets.QLabel()
        self.timer_label.setFont(mono_font())

        self.indicator = QtWidgets.QRadioButton()

        self.set_indicator()

        self.layout.addWidget(QtWidgets.QLabel("Uptime"), 0, 0, alignment=QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.layout.addWidget(self.timer_label, 1, 0)
        self.layout.addWidget(QtWidgets.QLabel("Status"), 0, 1)
        self.layout.addWidget(self.indicator, 1, 1)

        self.setFrameStyle(QtWidgets.QFrame.StyledPanel | QtWidgets.QFrame.Raised)


        self.setLayout(self.layout)

    def check_timeout(self):
        if (time.time() - self._last_heartbeat) > (self.timeout_dur/1000):
            self._state = True
            self.set_indicator("alarm")
            self.timeout.emit(True)
        else:
            self._state = False
            self.set_indicator("")

    def set_indicator(self, state=None):

        if state == 'alarm':
            self.setStyleSheet(styles.HEARTBEAT_ALARM)

        else:
            self.setStyleSheet(styles.HEARTBEAT_NORMAL)

    def start_timer(self, update_interval=None):
        """
        Args:
            update_interval (float): How often (in ms) the timer should be updated.
        """
        self.start_time = time.time()
        self._last_heartbeat = self.start_time

        if update_interval:
            self.update_interval = update_interval

        self.timer.start(self.update_interval)

    def stop_timer(self):
        """
        you can read the sign ya punk
        """
        self.timer.stop()
        self.setText("")

    @QtCore.Slot(float)
    def beatheart(self, heartbeat_time):
        self._last_heartbeat = heartbeat_time

    def _heartbeat(self):
        """
        Called every (update_interval) milliseconds to set the text of the timer.

        """
        current_time = time.time()
        self.heartbeat.emit(current_time)

        secs_elapsed = current_time-self.start_time
        self.timer_label.setText("{:02d}:{:02d}:{:.2f}".format(int(secs_elapsed/3600), int((secs_elapsed/60))%60, secs_elapsed%60))

        self.check_timeout()

class Power_Button(QtWidgets.QPushButton):

    def __init__(self):

        super(Power_Button, self).__init__()
        self.init_ui()

    def init_ui(self):
        pass




from .pigpio_mocks import patch_pigpio_base, patch_pigpio_gpio, soft_frequencies
from vent.io.devices.valves import OnOffValve, PWMControlValve, SimOnOffValve, SimControlValve
from secrets import token_bytes

import pytest
import random


@pytest.mark.parametrize("gpio", random.sample(range(31), 16))
@pytest.mark.parametrize("form", ['Normally Closed', 'Normally Open'])
def test_form(patch_pigpio_gpio, gpio, form):
    """__________________________________________________________________________________________________________TEST #1
     Tests the set/get interface of SolenoidBase/children
         - Initializes an OnOffValve
         - Sets a form
         - asserts form is set to expected form
    """
    valve = OnOffValve(gpio)
    valve.form = form
    assert valve.form == form
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("gpio", random.sample(range(31), 16))
@pytest.mark.parametrize("form", ['Normally Closed', 'Normally Open'])
def test_on_off_valve(patch_pigpio_gpio, gpio, form):
    """__________________________________________________________________________________________________________TEST #2
     Tests the open/close interface of an OnOffValve
         - Initializes an OnOffValve
         - records is_open
         - Opens the valve, records is_open
         - Closes the valve, records is_open
         - Asserts that the valve follows the pattern of a cycled valve with the set form
    """
    NC = 'Normally Closed'
    with pytest.raises(ValueError):
        OnOffValve(gpio, form='Not A Real Form')
    valve = OnOffValve(gpio, form)
    assert not valve.is_open if form == NC else valve.is_open
    valve.open()
    assert valve.is_open
    valve.close()
    assert not valve.is_open
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(31)])
def test_pwm_control_valve(patch_pigpio_gpio, seed):
    """__________________________________________________________________________________________________________TEST #3
     Tests the open/close interface of an PWMControlValve. This should mimic the behavior of an OnOffValve, but uses a
        different backend.
         - Tests that an exception is raised if an attempt is made to initialize w/o a response curve
         - Initializes a PWMControlValve
         - Checks proper functioning of open()/close() commands
         - Checks that valve.setpoint = 100 -> sets the highest duty of the 'rising' == True column of the response file
            - And that valve.is_open works
         - Checks that valve.setpoint = 1 -> sets the lowest duty from the 'rising' == False column of the response file
            - And that valve.is_open still reports that the valve is open
        - Checks that valve.setpoint = 0 -> valve.duty = 0 & valve.is_open reports false
    """
    random.seed(seed)
    gpio = random.choice(PWMControlValve._HARDWARE_PWM_PINS)
    form = 'Normally Closed'
    '''with pytest.raises(NotImplementedError):
        PWMControlValve(gpio, form)'''
    with pytest.raises(NotImplementedError):
        PWMControlValve(gpio, 'Normally Open')
    if random.getrandbits(1):
        valve = PWMControlValve(gpio, form)
    else:
        valve = PWMControlValve(gpio, form, response="vent/io/config/calibration/SMC_PVQ31_5G_23_01N_response")
    # Using standard On/Off commands:
    assert not valve.is_open
    valve.open()
    assert valve.is_open
    valve.close()
    assert not valve.is_open
    # Using setpoint:
    valve.setpoint = 100
    assert valve.is_open
    assert round(valve.duty, 4) == valve._response_array[:, 1].max()
    valve.setpoint = 1
    assert valve.is_open
    assert round(valve.duty, 4) == valve._response_array[valve._response_array[:, 2].nonzero(), 2].min()
    valve.setpoint = 0
    assert not valve.is_open
    assert valve.duty == 0
    with pytest.raises(ValueError):
        valve.setpoint = 101
    with pytest.raises(ValueError):
        valve.setpoint = -1
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("gpio", random.sample(range(31), 16))
@pytest.mark.parametrize("form", ['Normally Closed', 'Normally Open'])
def test_sim_on_off_valve(patch_pigpio_gpio, gpio, form):
    """__________________________________________________________________________________________________________TEST #4
     Tests the open/close interface of an SimOnOffValve
         - Initializes an OnOffValve
         - records is_open
         - Opens the valve, records is_open
         - Closes the valve, records is_open
         - Asserts that the valve follows the pattern of a cycled valve with the set form
    """
    NC = 'Normally Closed'
    valve = SimOnOffValve(gpio, form)
    assert not valve.is_open if form == NC else valve.is_open
    valve.open()
    assert valve.is_open
    valve.close()
    assert not valve.is_open
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("gpio", random.sample(range(31), 16))
def test_sim_control_valve(patch_pigpio_gpio, gpio):
    """__________________________________________________________________________________________________________TEST #5
     Tests SimControlValve in much the same way as PWMControlValve.
    """
    form = 'Normally Closed'
    with pytest.raises(NotImplementedError):
        SimControlValve(gpio, form, response="fake/path/to/file")
    with pytest.raises(NotImplementedError):
        SimControlValve(gpio, 'Normally Open')
    valve = SimControlValve(gpio, form)
    # Using standard On/Off commands:
    assert not valve.is_open
    valve.open()
    assert valve.is_open
    valve.close()
    assert not valve.is_open
    # Using setpoint:
    valve.setpoint = 100
    assert valve.is_open
    valve.setpoint = 1
    assert valve.is_open
    valve.setpoint = 0
    assert not valve.is_open
    with pytest.raises(ValueError):
        valve.setpoint = 101
    with pytest.raises(ValueError):
        valve.setpoint = -1
    """__________________________________________________________________________________________________________
    """


import pytest

import pdb
import time
import copy

import numpy as np

from vent.alarm import condition, ALARM_RULES, AlarmType, AlarmSeverity, Alarm, Alarm_Manager
from vent.alarm.rule import Alarm_Rule

from vent.common.values import ValueName, SENSOR
from vent.common.message import SensorValues

##########
# conditions

n_samples = 100


@pytest.fixture
def fake_sensors():
    def _fake_sensor(arg=None):
        # make an empty SensorValues
        vals = {k:0 for k in ValueName}
        vals.update({k:0 for k in SensorValues.additional_values})

        # since 0 is out of range for fio2, manually set it
        # FIXME: find values that by definition don't raise any of the default rules
        vals[ValueName.FIO2] = 80

        # update with any in kwargs
        if arg:
            for k, v in arg.items():
                vals[k] = v

        sensors = SensorValues(vals=vals)

        return sensors
    return _fake_sensor

@pytest.fixture
def fake_rule():
    def _fake_rule(alarm_type = AlarmType.HIGH_PRESSURE,
                   latch = False,
                   persistent = False,
                   conditions = None):

        if not conditions:
            conditions = (
                (
                    AlarmSeverity.LOW,
                    condition.ValueCondition(
                        value_name=ValueName.PRESSURE,
                        limit=1,
                        mode='max',
                    )
                ),
                (
                    AlarmSeverity.MEDIUM,
                    condition.ValueCondition(
                        value_name=ValueName.PRESSURE,
                        limit=2,
                        mode='max'
                    )
                ),
                (
                    AlarmSeverity.HIGH,
                    condition.ValueCondition(
                        value_name=ValueName.PRESSURE,
                        limit=3,
                        mode='max'
                    )
                )
            )
        rule = Alarm_Rule(
            name=alarm_type,
            latch=latch,
            persistent=persistent,
            conditions=conditions
        )

        return rule

    return _fake_rule

#############################
# conditions

@pytest.mark.parametrize("test_value", [k for k in SENSOR.keys()])
def test_value_condition(fake_sensors, test_value):


    for i in range(n_samples):
        # test min
        test_val = (np.random.rand()-0.5)*100
        min_cond = condition.ValueCondition(test_value, test_val, 'min')

        min_no_alarms = fake_sensors({test_value: test_val+1})
        min_yes_alarms = fake_sensors({test_value: test_val - 1})

        assert min_cond.check(min_no_alarms) == False
        assert min_cond.check(min_yes_alarms) == True

        # test max
        max_cond = condition.ValueCondition(test_value, test_val, 'max')

        max_no_alarms = fake_sensors({test_value: test_val - 1})
        max_yes_alarms = fake_sensors({test_value: test_val + 1})

        assert max_cond.check(max_no_alarms) == False
        assert max_cond.check(max_yes_alarms) == True

        # test that the @mode.setter works for already created objects
        min_cond.mode = 'max'
        max_cond.mode = 'min'

        assert min_cond.check(max_no_alarms) == False
        assert min_cond.check(max_yes_alarms) == True
        assert max_cond.check(min_no_alarms) == False
        assert max_cond.check(min_yes_alarms) == True

    # test that other values don't do anything
    other_cond = condition.ValueCondition(test_value, 1, 'max')

    for value in ValueName:
        if value == test_value:
            other_sensor = fake_sensors({value: 2})
        else:
            other_sensor = fake_sensors({value: 2, test_value: 0})

        assert other_sensor[value] == 2

        if value == test_value:
            assert other_cond.check(other_sensor) == True
        else:
            assert other_cond.check(other_sensor) == False

@pytest.mark.parametrize("test_value", [k for k in SENSOR.keys()])
def test_cyclevalue_condition(fake_sensors, test_value):

    for i in range(n_samples):
        n_cycles = np.random.randint(1, 100)

        cond = condition.CycleValueCondition(
            value_name=test_value,
            limit=1,
            mode='max',
            n_cycles=n_cycles
        )

        sensors = fake_sensors()

        # test that just checking alone without incrementing cycle doesn't trigger
        sensors[test_value] = 2

        for j in range(n_cycles * 2):
            assert cond.check(sensors) == False

        # test the straightforward case, goes out of range and stays out of range
        for j in range(n_cycles * 2):
            sensors.breath_count = j
            if j < n_cycles:
                assert cond.check(sensors) == False
            else:
                assert cond.check(sensors) == True

        # test that going under the limit resets the cycle count check
        sensors[test_value] = 0
        sensors.breath_count += 1
        assert cond.check(sensors) == False
        sensors[test_value] = 2
        sensors.breath_count += 1
        assert cond.check(sensors) == False

    # test discontinuous breath cycles
    # test that discontinuous checks without conflicting info still trigger
    cond = condition.CycleValueCondition(
            value_name=test_value,
            limit=1,
            mode='max',
            n_cycles=10
        )
    sensors = fake_sensors()
    sensors[test_value] = 2

    assert cond.check(sensors) == False
    sensors.breath_count = 11
    assert cond.check(sensors) == True

    # but that discontinuous checks with conflicting info dont trigger
    cond = condition.CycleValueCondition(
        value_name=test_value,
        limit=1,
        mode='max',
        n_cycles=10
    )
    sensors = fake_sensors()
    sensors[test_value] = 2

    assert cond.check(sensors) == False
    sensors.breath_count = 5
    sensors[test_value] = 0
    assert cond.check(sensors) == False
    sensors.breath_count = 11
    sensors[test_value] = 2
    assert cond.check(sensors) == False

def test_alarmseverity_condition():
    # FIXME
    pass

def test_cyclealarmseverity_condition():
    # FIXME
    pass

def test_condition_addition():
    # FIXME
    pass

def test_condition_dependency():
    # FIXME
    pass


###############################

# rules
#@pytest.mark.parametrize("alarm_rule", [k for k in ALARM_RULES.values()])
def test_alarm_rule(fake_sensors):
    """
    test the alarm rule class itself

    assume the individual conditions have been tested
    """

    rule = Alarm_Rule(
        name=AlarmType.HIGH_PRESSURE,
        latch=False,
        persistent=False,
        conditions=(
            (
                AlarmSeverity.LOW,
                condition.ValueCondition(
                    value_name=ValueName.PRESSURE,
                    limit=1,
                    mode='max',
                )
            ),
            (
                AlarmSeverity.MEDIUM,
                condition.ValueCondition(
                    value_name=ValueName.PRESSURE,
                    limit=2,
                    mode='max'
                ) + \
                condition.CycleAlarmSeverityCondition(
                    alarm_type = AlarmType.HIGH_PRESSURE,
                    severity   = AlarmSeverity.LOW,
                    n_cycles = 2
                )
            ),
            (
                AlarmSeverity.HIGH,
                condition.ValueCondition(
                    value_name=ValueName.PRESSURE,
                    limit=3,
                    mode='max'
                ) + \
                condition.CycleAlarmSeverityCondition(
                    alarm_type = AlarmType.HIGH_PRESSURE,
                    severity   = AlarmSeverity.MEDIUM,
                    n_cycles = 2
                )
            ),
        )
    )

    sensors = fake_sensors()

    # test that initial check is off
    assert rule.check(sensors) == AlarmSeverity.OFF

    # test low severity alarm
    sensors.PRESSURE = 1.5
    sensors.breath_count += 1

    assert rule.check(sensors) == AlarmSeverity.LOW

    # register alarm manually
    # (alarm should call register_alarm)
    low_alarm = Alarm(
        AlarmType.HIGH_PRESSURE,
        AlarmSeverity.LOW,
        latch=False
    )

    # test that we don't jump to medium
    sensors.PRESSURE = 2.5
    sensors.breath_count += 1

    assert rule.check(sensors) == AlarmSeverity.LOW

    # now check that we go to medium
    sensors.breath_count += 2

    assert rule.check(sensors) == AlarmSeverity.MEDIUM

    med_alarm = Alarm(
        AlarmType.HIGH_PRESSURE,
        AlarmSeverity.MEDIUM,
        latch=False
    )

    # keep at medium
    sensors.PRESSURE = 3.5
    sensors.breath_count += 1

    assert rule.check(sensors) == AlarmSeverity.MEDIUM

    sensors.breath_count += 2

    assert rule.check(sensors) == AlarmSeverity.HIGH

def test_rules_individual():
    # test that each individual rule does what we think it does
    # FIXME
    pass


##############################
#
def test_alarm_manager_raise(fake_sensors):
    """
    Test that the alarm manager raises a single alarm

    Args:
        fake_sensors:

    Returns:

    """

    manager = Alarm_Manager()
    manager.reset()
    assert len(manager.active_alarms) == 0

    # check that we got all the alarm rules
    for alarm_type, rule in ALARM_RULES.items():
        assert alarm_type in manager.rules.keys()

    # make callback to count emitted alarms
    global alarms_emitted
    alarms_emitted = 0
    def alarm_cb(alarm):
        assert isinstance(alarm, Alarm)
        global alarms_emitted
        alarms_emitted += 1

    manager.add_callback(alarm_cb)

    # take a value out of range and test that an alarm is raised an emitted
    sensor = fake_sensors()

    manager.update(sensor)
    assert len(manager.active_alarms) == 0
    assert alarms_emitted == 0

    sensor.PRESSURE = ALARM_RULES[AlarmType.HIGH_PRESSURE].conditions[0][1].limit + 1
    manager.update(sensor)
    assert AlarmType.HIGH_PRESSURE in manager.active_alarms.keys()
    assert alarms_emitted == 1

def test_alarm_manager_escalation(fake_sensors, fake_rule):
    """
    For unlatched alarms, test that alarms are:

    * emitted when alarm severity is raised
    * decremented when alarm severity is decreased and alarm is not latched
    * deactivated and deleted
    """

    manager = Alarm_Manager()
    manager.reset()

    assert len(manager.active_alarms) == 0

    #
    # create alarm rule for testing
    rule = fake_rule(latch=False, persistent=False)
    manager.load_rule(rule)

    assert manager.rules[rule.name] is rule

    # callback to catch emitted alarms
    global caught_alarm
    caught_alarm = None

    def catch_alarm(alarm):
        global caught_alarm
        caught_alarm = alarm

    manager.add_callback(catch_alarm)

    # raise and see that we get a new alarm
    sensors = fake_sensors()
    sensors.PRESSURE = 1.5

    manager.update(sensors)

    assert isinstance(caught_alarm, Alarm)
    assert caught_alarm.severity == AlarmSeverity.LOW
    assert manager.active_alarms[rule.name] is caught_alarm

    # raise to medium
    sensors.PRESSURE = 2.5
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.MEDIUM
    assert manager.active_alarms[rule.name] is caught_alarm

    # then high
    sensors.PRESSURE = 3.5
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert manager.active_alarms[rule.name] is caught_alarm

    # now decrement
    sensors.PRESSURE = 2.5
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.MEDIUM
    assert manager.active_alarms[rule.name] is caught_alarm

    sensors.PRESSURE = 1.5
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.LOW
    assert manager.active_alarms[rule.name] is caught_alarm

    # when alarm is brought back into safe range, should be deleted from active_alarms
    sensors.PRESSURE = 0.5
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.OFF
    assert rule.name not in manager.active_alarms.keys()
    assert len(manager.active_alarms) == 0



def test_alarm_manager_latch(fake_sensors, fake_rule):
    """
    * not decremented when alarm severity is decreased and alarm is latched
    * dismissed when latched *only when* alarm condition is OFF AND has been dismissed

    Args:
        fake_sensors:
        fake_rule:

    Returns:

    """

    manager = Alarm_Manager()
    manager.reset()

    assert len(manager.active_alarms) == 0

    # create alarm rule for testing
    rule = fake_rule(latch=True, persistent=False)
    manager.load_rule(rule)
    assert manager.rules[rule.name] is rule

    # callback to catch emitted alarms
    global caught_alarm
    global n_alarms
    n_alarms = 0
    caught_alarm = None

    def catch_alarm(alarm):
        global caught_alarm
        global n_alarms
        caught_alarm = alarm
        n_alarms += 1

    manager.add_callback(catch_alarm)

    # raise and see that we get a new alarm
    sensors = fake_sensors()

    # raise alarm to HIGH
    sensors.PRESSURE = 3.5
    manager.update(sensors)

    assert isinstance(caught_alarm, Alarm)
    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert manager.active_alarms[rule.name] is caught_alarm
    assert n_alarms == 1

    # raise alarm range to OFF, but don't request dismiss
    sensors.PRESSURE = 0.5
    manager.update(sensors)

    # no alarm should have been emitted
    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert manager.active_alarms[rule.name] is caught_alarm
    assert n_alarms == 1

    # raise back to HIGH
    sensors.PRESSURE = 3.5
    manager.update(sensors)

    # dismiss but only drop to MEDIUM, so alarm should not be emitted
    manager.dismiss_alarm(rule.name)

    assert rule.name in manager.pending_clears

    sensors.PRESSURE = 2.5
    manager.update(sensors)

    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert manager.active_alarms[rule.name] is caught_alarm
    assert n_alarms == 1

    # dropping to OFF should finally clear the alarm
    sensors.PRESSURE = 0.5
    manager.update(sensors)

    assert caught_alarm.severity == AlarmSeverity.OFF
    assert rule.name not in manager.active_alarms.keys()
    assert len(manager.active_alarms) == 0
    assert len(manager.pending_clears) == 0
    assert n_alarms == 2


def test_alarm_manager_dismiss(fake_sensors, fake_rule):
    """
    * dismissed when not latched until alarm conditions return to off then back on
    * dismissed when not latched until duration regardless of alarm condition

    Args:
        fake_sensors:
        fake_rule:

    Returns:

    """
    manager = Alarm_Manager()
    manager.reset()

    assert len(manager.active_alarms) == 0

    # create alarm rule for testing
    rule = fake_rule(latch=False, persistent=False)
    manager.load_rule(rule)
    assert manager.rules[rule.name] is rule

    # callback to catch emitted alarms
    global caught_alarm
    global n_alarms
    n_alarms = 0
    caught_alarm = None

    def catch_alarm(alarm):
        global caught_alarm
        global n_alarms
        caught_alarm = alarm
        n_alarms += 1

    manager.add_callback(catch_alarm)

    # raise and see that we get a new alarm
    sensors = fake_sensors()
    sensors.PRESSURE = 3.5
    manager.update(sensors)

    assert isinstance(caught_alarm, Alarm)
    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert n_alarms == 1

    # dismiss the alarm, make sure it's gone
    manager.dismiss_alarm(rule.name)
    assert rule.name in manager.cleared_alarms
    assert caught_alarm.severity == AlarmSeverity.OFF
    assert n_alarms == 2

    # update with same sensor values, make sure alarm isn't emitted
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.OFF
    assert n_alarms == 2

    # take down to OFF range and back up, should get alarm
    sensors.PRESSURE = 0.5
    manager.update(sensors)

    assert rule.name not in manager.cleared_alarms
    assert n_alarms == 2

    sensors.PRESSURE = 3.5
    manager.update(sensors)

    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert n_alarms == 3

    #####################
    # test timed dismissal
    dismiss_time = 2
    manager.dismiss_alarm(rule.name, duration = dismiss_time)

    assert rule.name in manager.snoozed_alarms.keys()
    assert caught_alarm.severity == AlarmSeverity.OFF
    assert n_alarms ==4

    # stash snooze time to check we're not going over
    snooze_time = manager.snoozed_alarms[rule.name]
    # make sure it's not some crazy value
    assert snooze_time <= time.time() + dismiss_time

    # give flip off and on again, check that alarm is not emitted
    # as would be typical
    sensors.PRESSURE = 0.5
    manager.update(sensors)
    sensors.PRESSURE = 3.5
    manager.update(sensors)

    assert caught_alarm.severity == AlarmSeverity.OFF
    assert n_alarms == 4
    assert time.time() < snooze_time

    while time.time() < snooze_time:
        manager.update(sensors)
        assert caught_alarm.severity == AlarmSeverity.OFF
        assert n_alarms == 4
        time.sleep(0.1)

    # now that time has passed, should get an alarm
    assert time.time() > snooze_time
    manager.update(sensors)
    assert caught_alarm.severity == AlarmSeverity.HIGH
    assert n_alarms == 5


def test_alarm_manager_logged_alarms(fake_sensors, fake_rule):
    manager = Alarm_Manager()
    manager.reset()

    assert len(manager.active_alarms) == 0

    # create alarm rule for testing
    rule = fake_rule(latch=False, persistent=False)
    manager.load_rule(rule)
    assert manager.rules[rule.name] is rule

    # callback to catch emitted alarms
    global caught_alarm
    global n_alarms
    n_alarms = 0
    caught_alarm = None

    def catch_alarm(alarm):
        global caught_alarm
        global n_alarms
        caught_alarm = alarm
        n_alarms += 1

    manager.add_callback(catch_alarm)

    # raise and see that we get a new alarm
    sensors = fake_sensors()

    # raise alarm, catch it, turn alarm off, check that it's put in logged_alarms
    sensors.PRESSURE = 3.5
    manager.update(sensors)
    assert isinstance(caught_alarm, Alarm)
    assert caught_alarm.severity == AlarmSeverity.HIGH

    high_alarm = caught_alarm

    sensors.PRESSURE = 0.5
    manager.update(sensors)
    assert high_alarm in manager.logged_alarms
    assert high_alarm not in manager.active_alarms.values()






from collections import OrderedDict as odict
from enum import Enum, auto

# TODO: Zhenyu's job is to make sure the print value is an intepretable string
class ValueName(Enum):
    #Setting that are likely important for future adjustements
    PIP = auto()       # PIP pressure
    PIP_TIME = auto()  # time to reach PIP
    PEEP = auto()      # PEEP pressure
    PEEP_TIME = auto() # time to reach PEEP
    BREATHS_PER_MINUTE = auto()
    INSPIRATION_TIME_SEC = auto()
    IE_RATIO = auto()
    #Settings that are read out, but can not be controlled by software
    FIO2 = auto()
    TEMP = auto()
    HUMIDITY = auto()
    VTE = auto()
    PRESSURE = auto()

controllable_values = {
    ValueName.PIP,
    ValueName.PIP_TIME,
    ValueName.PEEP,
    ValueName.BREATHS_PER_MINUTE,
    ValueName.INSPIRATION_TIME_SEC,
    ValueName.IE_RATIO
}

non_controllable_values = {
    # TODO: is PEEP_TIME controllable? control_module doesn't allow set_control on it
    ValueName.PEEP_TIME,
    ValueName.FIO2,
    ValueName.TEMP,
    ValueName.HUMIDITY,
    ValueName.VTE,
    ValueName.PRESSURE,
}


class Value(object):

    def __init__(self,
                 name: str,
                 units: str,
                 abs_range: tuple,
                 safe_range: tuple,
                 decimals: int,
                 control: bool,
                 sensor: bool,
                 default: (int, float) = None):
        """
        Definition of a value.

        Used by the GUI and control module to set defaults.

        Args:
            name (str): Human-readable name of the value
            units (str): Human-readable description of units
            abs_range (tuple): tuple of ints or floats setting the logical limit of the value,
                eg. a percent between 0 and 100, (0, 100)
            safe_range (tuple): tuple of ints or floats setting the safe ranges of the value,

                note::

                    this is not the same thing as the user-set alarm values,
                    though the user-set alarm values are initialized as ``safe_range``.

            decimals (int): the number of decimals of precision used when displaying the value
        """

        self._name = None
        self._units = None
        self._abs_range = None
        self._safe_range = None
        self._decimals = None
        self._default = None
        self._control = None
        self._sensor = None

        self.name = name
        self.units = units
        self.abs_range = abs_range
        self.safe_range = safe_range
        self.decimals = decimals
        self.control = control
        self.sensor = sensor

        if default is not None:
            self.default = default

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name):
        assert(isinstance(name, str))
        self._name = name

    @property
    def abs_range(self) -> tuple:
        return self._abs_range

    @abs_range.setter
    def abs_range(self, abs_range):
        assert(isinstance(abs_range, tuple) or isinstance(abs_range, list))
        assert(all([isinstance(x, int) or isinstance(x, float) for x in abs_range]))
        self._abs_range = abs_range

    @property
    def safe_range(self) -> tuple:
        return self._safe_range

    @safe_range.setter
    def safe_range(self, safe_range):
        assert(isinstance(safe_range, tuple) or isinstance(safe_range, list))
        assert(all([isinstance(x, int) or isinstance(x, float) for x in safe_range]))
        self._safe_range = safe_range

    @property
    def decimals(self) -> int:
        return self._decimals

    @decimals.setter
    def decimals(self, decimals):
        assert(isinstance(decimals, int))
        self._decimals = decimals

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, default):
        assert(isinstance(default, int) or isinstance(default, float))
        self._default = default

    @property
    def control(self):
        return self._control

    @control.setter
    def control(self, control):
        assert(isinstance(control, bool))
        self._control = control

    @property
    def sensor(self):
        return self._sensor

    @sensor.setter
    def sensor(self, sensor):
        assert(isinstance(sensor, bool))
        self._sensor = sensor

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return self.__getattribute__(key)

    def to_dict(self):
        return {
            'name': self.name,
            'units': self.units,
            'abs_range': self.abs_range,
            'safe_range': self.safe_range,
            'decimals': self.decimals,
            'default': self.default,
            'control': self.control,
            'sensor': self.sensor
                }




VALUES = odict({
    ValueName.FIO2: Value(**{ 'name': 'FiO2',
        'units': '%',
        'abs_range': (0, 100),
        'safe_range': (20, 100),
        'decimals' : 1,
        'control': False,
        'sensor': True
    }),
    ValueName.TEMP: Value(**{
        'name': 'Temp',
        'units': '\N{DEGREE SIGN}C',
        'abs_range': (35, 40),
        'safe_range': (36, 39),
        'decimals': 1,
        'control': False,
        'sensor': True
    }),
    ValueName.HUMIDITY: Value(**{
        'name': 'Humidity',
        'units': '%',
        'abs_range': (0, 100),
        'safe_range': (70, 100),
        'decimals': 1,
        'control': False,
        'sensor': True
    }),
    ValueName.VTE: Value(**{
        'name': 'VTE',
        'units': 'l',  # Unit is liters :-)
        'abs_range': (0, 100),
        'safe_range': (0, 100),
        'decimals': 2,
        'control': False,
        'sensor': True
    }),
    ValueName.PRESSURE: Value(**{
        'name': 'Pressure',
        'units': 'mmH2O',
        'abs_range': (0,70),
        'safe_range': (0,60),
        'decimals': 1,
        'control': False,
        'sensor': True
    }),
    ValueName.IE_RATIO: Value(**{
        'name': 'I:E Ratio',
        'units': '',
        'abs_range': (0, 2),
        'safe_range': (0.33, 1),
        'decimals': 2,
        'control': False,
        'sensor': False
    }),
    ValueName.PIP: Value(**{
        'name': 'PIP', # (Peak Inspiratory Pressure)
        'units': 'cm H2O',
        'abs_range': (0, 70), # FIXME
        'safe_range': (0, 50), # From DrDan https://tigervents.slack.com/archives/C011MRVJS7L/p1588190130492300
        'default': 22,           # FIXME
        'decimals': 1,
        'control': True,
        'sensor': True
    }),
    ValueName.PIP_TIME: Value(**{
        'name': 'PIPt',
        'units': 'seconds',
        'abs_range': (0, 5),  # FIXME
        'safe_range': (0.2, 0.5),  # FIXME
        'default': 0.3,  # FIXME
        'decimals': 1,
        'control': True,
        'sensor': False
    }),
    ValueName.INSPIRATION_TIME_SEC: Value(**{
        'name': 'INSPt',
        'units': 'seconds',
        'abs_range': (0, 5),  # FIXME
        'safe_range': (1, 3.0),  # FIXME
        'default': 2.0,  # FIXME
        'decimals': 1,
        'control': True,
        'sensor': True
    }),
    ValueName.PEEP: Value(**{
        'name': 'PEEP', #  (Positive End Expiratory Pressure)
        'units': 'cm H2O',
        'abs_range': (0, 20),  # FIXME
        'safe_range': (0, 16), # From DrDan https://tigervents.slack.com/archives/C011MRVJS7L/p1588190130492300
        'default': 5,            # FIXME
        'decimals': 1,
        'control': True,
        'sensor': True
    }),
    ValueName.PEEP_TIME: Value(**{
        'name': 'PEEPt',
        'units': 'seconds',
        'abs_range': (0, 2),  # FIXME
        'safe_range': (0, 1.0),  # FIXME
        'default': 0.5,  # FIXME
        'decimals': 1,
        'control': True,
        'sensor': False
    }),
    ValueName.BREATHS_PER_MINUTE: Value(**{
        'name': 'RR', # Daniel re: FDA labels
        'units': 'BPM', # Daniel re: FDA labels
        'abs_range': (0, 50), # FIXME
        'safe_range': (10, 30), # Stanford's socs https://www.vent4us.org/technical
        'default': 17,            # FIXME
        'decimals': 1,
        'control': True,
        'sensor': True
    }),
})

SENSOR = odict({
    k:v for k, v in VALUES.items() if v.sensor == True
})
"""
Values to monitor but not control. 

Used to set alarms for out-of-bounds sensor values. These should be sent from the control module and not computed.::

    {
        'name' (str):  Human readable name,
        'units' (str): units string, (like degrees or %),
        'abs_range' (tuple): absolute possible range of values,
        'safe_range' (tuple): range outside of which a warning will be raised,
        'decimals' (int): The number of decimals of precision this number should be displayed with
    }
"""


CONTROL = odict({
    k: v for k, v in VALUES.items() if v.control == True
})
"""
Values to control but not monitor.

Sent to control module to control operation of ventilator.::

    {
        'name' (str):  Human readable name,
        'units' (str): units string, (like degrees or %),
        'abs_range' (tuple): absolute possible range of values,
        'safe_range' (tuple): range outside of which a warning will be raised,
        'default' (int, float): the default value of the parameter,
        'decimals' (int): The number of decimals of precision this number should be displayed with
    }
"""

LIMITS = {

}
"""
Values that are dependent on other values::

    {
        "dependent_value": (
            ['value_1', 'value_2'],
            callable_returning_boolean
        }
    }
    
Where the first argument in the tuple is a list of the values that will be 
given as argument to the ``callable_returning_boolean`` which will return
whether (``True``) or not (``False``) a value is allowed.
"""
